"""
Core request classes that implement calls available in all versions of RWS.

RWSRequest subclasses represent URL endpoints. They can be passed to a RWSConnection get or post method
as appropriate.

TODO: Note that might want to make the Request objects responsible for deciding whether they are get, post, patch etc.

"""
import datetime

from rwslib.rwsobjects import (
    RWSResponse,
    RWSStudies,
    RWSStudyMetadataVersions,
    RWSSubjects,
    RWSPostResponse,
)
from six.moves.urllib_parse import urlencode


# -----------------------------------------------------------------------------------------------------------------------
# Function utilities I don't want to muddy class heirarchy with


def check_dataset_type(dataset_type):
    """
    Datasets may only be regular or raw
    :type dataset_type: str
    """
    if dataset_type.lower() not in ["regular", "raw"]:
        raise ValueError("Dataset type not 'regular' or 'raw' is %s" % dataset_type)


def format_date_argument(date_element):
    """
    Take a date as either a datetime.date/datetime or a string and return it as a iso8601 formatted value
    :param Union[datetime.date, datetime.datetime] date_element: passed argument
    :rtype str
    :return:
    """
    if not isinstance(date_element, (datetime.datetime, datetime.date)):
        # TODO:
        if "T" in date_element:
            _date = datetime.datetime.strptime(date_element, "%Y-%m-%dT%H:%M:%S")
        else:
            _date = datetime.datetime.strptime(date_element, "%Y-%m-%d").date()
    else:
        _date = date_element
    return _date.isoformat()


# -------------------------------------------------------------------------------------------------------
# Utility functions
#
def make_url(*args, **kwargs):
    """Makes a URL from component parts"""
    base = "/".join(args)
    if kwargs:
        return "%s?%s" % (base, urlencode(kwargs))
    else:
        return base


class RWSRequest(object):
    """Base class for all RWS Requests"""

    requires_authorization = False
    method = "GET"  # Default

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def result(self, response):
        """
        Process a result to create a custom output
        :param requests.models.Response response: returned response
        :return:
        """
        # By default return text
        return response.text

    def url_path(self):
        """Return url path list"""
        raise NotImplementedError(
            "Override url_path in descendants of RWSRequest"
        )  # pragma: no cover

    def args(self):
        """Return additional args here as dict"""
        return {}

    def make_url(self, *args, **kwargs):
        # Note: Including this in the class as a convenience so that you can get all you need from a RWSObject rather
        # than having to import make_url as an additional import.
        return make_url(*args, **kwargs)


# -----------------------------------------------------------------------------------------------------------------------
# Useful subclasses
# -----------------------------------------------------------------------------------------------------------------------


class RWSGetRequest(RWSRequest):
    method = "GET"


class RWSAuthorizedGetRequest(RWSGetRequest):
    requires_authorization = True


class RWSPostRequest(RWSRequest):
    method = "POST"


class RWSAuthorizedPostRequest(RWSPostRequest):
    requires_authorization = True


class QueryOptionGetRequest(RWSAuthorizedGetRequest):
    """Manages requests that have known query string options"""

    KNOWN_QUERY_OPTIONS = []

    def _querystring(self):
        """Get additional keyword arguments"""

        kw = {}
        for key in self.KNOWN_QUERY_OPTIONS:
            val = getattr(self, key)
            if val is not None:
                kw[key] = val
        return kw


# ---------------------------------------------------------------------------------------------------------------------
# Implementations. These are all standards that have existed for a long time.
# ---------------------------------------------------------------------------------------------------------------------


class VersionRequest(RWSGetRequest):
    """Get RWS Version number"""

    def url_path(self):
        return make_url("version")


class BuildVersionRequest(RWSGetRequest):
    """Return the RWS build version number"""

    def url_path(self):
        return make_url("version", "build")


class CodeNameRequest(RWSGetRequest):
    """Return the RWS version codename"""

    def url_path(self):
        return make_url("version", "codename")


class DiagnosticsRequest(RWSGetRequest):
    """Return the RWS build version number"""

    def url_path(self):
        return make_url("diagnostics")


class TwoHundredRequest(RWSGetRequest):
    """Return RWS MAuth information"""

    def url_path(self):
        return make_url("twohundred")


class CacheFlushRequest(RWSAuthorizedGetRequest):
    """Calls RWS cache-flush"""

    def url_path(self):
        return make_url("webservice.aspx?CacheFlush")

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSResponse(response.text)


class ClinicalStudiesRequest(RWSAuthorizedGetRequest):
    """Return the list of clinical studies as a RWSStudies object.
       Clinical studies are the studies that you have access to as an EDC user.
    """

    def url_path(self):
        return make_url("studies")

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSStudies(response.text)


# ----------------------------------------------------------------------------------------------------------------------
# Base request classes for study versions (could also be used for drafts if ever implemented by RWS)
# ----------------------------------------------------------------------------------------------------------------------


class VersionRequestBase(RWSAuthorizedGetRequest):
    """Base class for study and library metadata version requests"""

    def __init__(self, project_name, oid):
        """
        :param str project_name: Project Name
        :param int oid: Metadata version OID (CRF Draft Number)
        """
        self.project_name = project_name
        self._oid = None  # Oid is a version OID, an integer identifier
        self.oid = oid

    @property
    def oid(self):
        return self._oid

    @oid.setter
    def oid(self, value):
        try:
            int(value)
        except ValueError:
            raise ValueError("oid must be an integer")
        self._oid = value

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return response.text


# -----------------------------------------------------------------------------------------------------------------------
# Related to Architect Studies and their drafts/versions
# -----------------------------------------------------------------------------------------------------------------------


class MetadataStudiesRequest(RWSAuthorizedGetRequest):
    """Return the list of metadata studies as a RWSStudies object.
       metadata_studies are the list of studies that you have access to as an
       Architect user.
    """

    def url_path(self):
        return make_url("metadata", "studies")

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSStudies(response.text)


class StudyDraftsRequest(RWSAuthorizedGetRequest):
    """Return the list of study drafts"""

    def __init__(self, project_name):
        """
        :param str project_name: Project Name
        """
        self.project_name = project_name

    def url_path(self):
        return make_url("metadata", "studies", self.project_name, "drafts")

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSStudyMetadataVersions(response.text)


class StudyVersionsRequest(RWSAuthorizedGetRequest):
    """Return the list of study versions"""

    def __init__(self, project_name):
        """
        :param str project_name: Project Name
        """
        self.project_name = project_name

    def url_path(self):
        return make_url("metadata", "studies", self.project_name, "versions")

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSStudyMetadataVersions(response.text)


class StudyVersionRequest(VersionRequestBase):
    """Return a study version as a string"""

    def url_path(self):
        return make_url(
            "metadata", "studies", self.project_name, "versions", str(self._oid)
        )


# NOTE: There is no StudyDraftRequest, this is something of an omission since you can list them...


# -------------------------------------------------------------------------------------------------
# Related to Architect Global Libraries and their drafts/versions
# -------------------------------------------------------------------------------------------------


class GlobalLibrariesRequest(RWSAuthorizedGetRequest):
    """Return the list of global libraries as a RWSStudies object.
       metadata_libraries are the list of libraries that you have access to as an
       Architect Global Library Volume user
    """

    def url_path(self):
        return make_url("metadata", "libraries")

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSStudies(response.text)


class GlobalLibraryDraftsRequest(RWSAuthorizedGetRequest):
    """Return the list of global library drafts"""

    def __init__(self, project_name):
        """
        :param str project_name: Project Name
        """
        self.project_name = project_name

    def url_path(self):
        return make_url("metadata", "libraries", self.project_name, "drafts")

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSStudyMetadataVersions(response.text)


class GlobalLibraryVersionsRequest(RWSAuthorizedGetRequest):
    """Return the list of global library versions"""

    def __init__(self, project_name):
        """
        :param str project_name: Project Name
        """
        self.project_name = project_name

    def url_path(self):
        return make_url("metadata", "libraries", self.project_name, "versions")

    def result(self, response):
        """
        Return RWSResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSStudyMetadataVersions(response.text)


class GlobalLibraryVersionRequest(VersionRequestBase):
    """Return a global library version as a string"""

    def url_path(self):
        return make_url(
            "metadata", "libraries", self.project_name, "versions", str(self._oid)
        )


class PostMetadataRequest(RWSAuthorizedPostRequest):
    """Post an ODM data transaction to Rave, get back an RWSResponse object"""

    def __init__(self, project_name, data, headers={"Content-type": "text/xml"}):
        """
        :param str project_name: Project Name
        :param str data: Data to dispatch
        :param dict headers: Headers to pass to client
        """
        self.project_name = project_name
        self.data = data
        self.headers = headers

    def args(self):
        """Return additional args here as dict (only for post data requests)"""
        kw = {"data": self.data, "headers": self.headers}
        return kw

    def url_path(self):
        return make_url("metadata", "studies", self.project_name, "drafts")

    def result(self, response):
        """
        Return RWSPostResponse object for success
        :param requests.models.Response response: request respnse
        """
        return RWSPostResponse(response.text)


# -------------------------------------------------------------------------------------------------
# Subject related
# -------------------------------------------------------------------------------------------------


class StudySubjectsRequest(RWSAuthorizedGetRequest):
    """
    Return the list of study subjects, defaults to the PROD environment
    """

    SUBJECT_KEY_TYPES = ["SubjectName", "SubjectUUID"]
    INCLUDE_OPTIONS = ["inactive", "deleted", "inactiveAndDeleted"]

    def __init__(
        self,
        project_name,
        environment_name,
        status=False,
        include=None,
        subject_key_type="SubjectName",
        links=False,
    ):
        """
        :param str project_name: Project/Study Name
        :param str environment_name: Study Environment Name (eg `Prod`)
        :param bool status: Add subject level workflow status
        :param str include: Query option to add to `include` parameter (see INCLUDE_OPTIONS for allowed values)
        :param str subject_key_type: Type of SubjectKey to have in the response (one of `SubjectName` or `SubjectUUID`)
        :param bool links: Add Deep Links to Output ODM

        .. note::
            If status == True then ?status=all
            If include then include parameter is also added to query string

        """

        self.project_name = project_name
        self.environment_name = environment_name
        self.status = status
        self.links = links
        self.include = None
        self.subject_key_type = subject_key_type
        # make sure the value for SubjectKeyType makes sense.
        if self.subject_key_type not in self.SUBJECT_KEY_TYPES:
            raise ValueError(
                "SubjectKeyType {} is not a valid value".format(self.subject_key_type)
            )

        if include is not None:
            if include not in self.INCLUDE_OPTIONS:
                raise ValueError(
                    "If provided, included must be one of %s"
                    % ",".join(self.INCLUDE_OPTIONS)
                )
        self.include = include

    def _querystring(self):
        """Additional keyword arguments"""
        kw = {}
        if self.status:
            kw["status"] = "all"

        if self.links:
            kw["links"] = "all"

        if self.include is not None:
            kw["include"] = self.include
        if self.subject_key_type != "SubjectName":
            kw["subjectKeyType"] = self.subject_key_type
        return kw

    def studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name)

    def url_path(self):
        return make_url(
            "studies", self.studyname_environment(), "subjects", **self._querystring()
        )

    def result(self, response):
        """
        Return RWSSubjects object for success
        :param requests.models.Response response: request response
        """
        response.encoding = "utf-8-sig"
        return RWSSubjects(response.text)


class PostDataRequest(RWSAuthorizedPostRequest):
    """Post an ODM data transaction to Rave, get back an RWSResponse object"""

    def __init__(self, data, headers={"Content-type": "text/xml"}):
        """
        Post an ODM to a RWS endpoint
        :param bytes data: Data to POST
        :param dict headers: Headers to add to request
        """
        self.data = data
        self.headers = headers

    def args(self):
        """Return additional args here as dict (only for post data requests)"""
        kw = {"data": self.data, "headers": self.headers}
        return kw

    def url_path(self):
        """
        Generate the Target Endpoint
        :return: Generated Endpoint
        """
        return make_url("webservice.aspx?PostODMClinicalData")

    def result(self, response):
        """
        Return RWSPostResponse object for success
        :param requests.models.Response response: request response
        """
        return RWSPostResponse(response.text)


# -------------------------------------------------------------------------------------------------
# ODM Clinical Data Datasets
# -------------------------------------------------------------------------------------------------


class ODMDatasetBase(RWSAuthorizedGetRequest):
    KNOWN_QUERY_OPTIONS = [
        "versionitem",
        "rawsuffix",
        "codelistsuffix",
        "decodesuffix",
        "stdsuffix",
        "start",
    ]

    def checkParams(self):
        check_dataset_type(self.dataset_type)
        # https://bitbucket.org/micktwomey/pyiso8601

    def _querystring(self):
        """Additional keyword arguments"""

        kw = {}

        for key in self.KNOWN_QUERY_OPTIONS:
            val = getattr(self, key)
            if val is not None:
                if key == "start":
                    # special case the date formatting
                    # this will raise a ValueError
                    # if the date format is a string and not one of
                    # YYYY-mm-ddTHH:MM:SS or YYYY-mm-dd
                    kw[key] = format_date_argument(val)
                else:
                    kw[key] = val
        return kw

    def _studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name)


class StudyDatasetRequest(ODMDatasetBase):
    """Return the text of the full datasets listing as an ODM string."""

    def __init__(
        self,
        project_name,
        environment_name,
        dataset_type="regular",
        start=None,
        rawsuffix=None,
        formoid=None,
        versionitem=None,
        codelistsuffix=None,
        decodesuffix=None,
        stdsuffix=None,
    ):
        """
        :param str project_name: Name of the Rave Study
        :param str environment_name: Name of the Rave Study Environment
        :param str dataset_type: Dataset type, one of 'regular' or 'raw'
        :param str start: Start date for the Report (iso8601)
        :param str rawsuffix: Adds raw data values to a full or incremental dataset, and identifies these
            values with {rawsuffix}.
        :param str formoid: FormOID for which to report
        :param str versionitem: Adds the subject's CRF version to the dataset, and identifies it with {version-item}
        :param str codelistsuffix: Adds code list OIDS for fields that use a code list to the dataset,
            and identifies these values with {codelistsuffix}.
        :param str decodesuffix: Add decoded values of items that have an associated code list to the dataset,
            and identifies these values with {decodesuffix}.
        :param str stdsuffix: Adds standard data values to a full or incremental dataset, and identifies these
            values with {stdsuffix}.
        """
        self.project_name = project_name
        self.environment_name = environment_name

        self.dataset_type = dataset_type.lower()
        self.formoid = formoid

        self.rawsuffix = rawsuffix
        self.start = start
        self.versionitem = versionitem
        self.codelistsuffix = codelistsuffix
        self.decodesuffix = decodesuffix
        self.stdsuffix = stdsuffix

        self.checkParams()

    def url_path(self):
        args = ["studies", self._studyname_environment(), "datasets", self.dataset_type]
        if self.formoid is not None:
            args.append(self.formoid)

        return make_url(*args, **self._querystring())


class VersionDatasetRequest(ODMDatasetBase):
    """
    Return the text of the full datasets for a version as an ODM string.
    By supplying formoid, will be filtered to just that formoid data
    """

    def __init__(
        self,
        project_name,
        environment_name,
        version_oid,
        dataset_type="regular",
        start=None,
        rawsuffix=None,
        formoid=None,
        versionitem=None,
        codelistsuffix=None,
        decodesuffix=None,
        stdsuffix=None,
    ):
        """
        :param str project_name: Name of the Rave Study
        :param str environment_name: Name of the Rave Study Environment
        :param str dataset_type: Dataset type, one of 'regular' or 'raw'
        :param str start: Start date for the Report (iso8601)
        :param str rawsuffix: Adds raw data values to a full or incremental dataset, and identifies these
            values with {rawsuffix}.
        :param str formoid: FormOID for which to report
        :param str versionitem: Adds the subject's CRF version to the dataset, and identifies it with {version-item}
        :param str codelistsuffix: Adds code list OIDS for fields that use a code list to the dataset,
            and identifies these values with {codelistsuffix}.
        :param str decodesuffix: Add decoded values of items that have an associated code list to the dataset,
            and identifies these values with {decodesuffix}.
        :param str stdsuffix: Adds standard data values to a full or incremental dataset, and identifies these
            values with {stdsuffix}.
        :param str version_oid: version for which to report
        """
        self.project_name = project_name
        self.environment_name = environment_name
        self.version_oid = version_oid

        self.dataset_type = dataset_type.lower()
        self.formoid = formoid

        self.rawsuffix = rawsuffix
        self.start = start
        self.versionitem = versionitem
        self.codelistsuffix = codelistsuffix
        self.decodesuffix = decodesuffix
        self.stdsuffix = stdsuffix

        self.checkParams()

    def url_path(self):
        args = [
            "studies",
            self._studyname_environment(),
            "versions",
            str(self.version_oid),
            "datasets",
            self.dataset_type,
        ]
        if self.formoid is not None:
            args.append(self.formoid)

        return make_url(*args, **self._querystring())


class SubjectDatasetRequest(ODMDatasetBase):
    """
    Return the text of the full datasets for a version as an ODM string.
    By supplying formoid, will be filtered to just that formoid data
    """

    def __init__(
        self,
        project_name,
        environment_name,
        subjectkey,
        dataset_type="regular",
        start=None,
        rawsuffix=None,
        formoid=None,
        versionitem=None,
        codelistsuffix=None,
        decodesuffix=None,
        stdsuffix=None,
    ):
        """
        :param str project_name: Name of the Rave Study
        :param str environment_name: Name of the Rave Study Environment
        :param str dataset_type: Dataset type, one of 'regular' or 'raw'
        :param str start: Start date for the Report (iso8601)
        :param str rawsuffix: Adds raw data values to a full or incremental dataset, and identifies these
            values with {rawsuffix}.
        :param str formoid: FormOID for which to report
        :param str versionitem: Adds the subject's CRF version to the dataset, and identifies it with {version-item}
        :param str codelistsuffix: Adds code list OIDS for fields that use a code list to the dataset,
            and identifies these values with {codelistsuffix}.
        :param str decodesuffix: Add decoded values of items that have an associated code list to the dataset,
            and identifies these values with {decodesuffix}.
        :param str stdsuffix: Adds standard data values to a full or incremental dataset, and identifies these
            values with {stdsuffix}.
        :param str subjectkey: Subject Key for subject for which to report
        """
        self.project_name = project_name
        self.environment_name = environment_name
        self.subjectkey = subjectkey

        self.dataset_type = dataset_type.lower()
        self.formoid = formoid

        self.rawsuffix = rawsuffix
        self.start = start
        self.versionitem = versionitem
        self.codelistsuffix = codelistsuffix
        self.decodesuffix = decodesuffix
        self.stdsuffix = stdsuffix

        self.checkParams()

    def url_path(self):
        args = [
            "studies",
            self._studyname_environment(),
            "subjects",
            str(self.subjectkey),
            "datasets",
            self.dataset_type,
        ]
        if self.formoid is not None:
            args.append(self.formoid)

        return make_url(*args, **self._querystring())


class ConfigurableDatasetRequest(RWSAuthorizedGetRequest):
    VALID_DATASET_FORMATS = ()

    def __init__(self, dataset_name, dataset_format="", params={}):
        """
        Create a new Configurable Dataset Request
        :param dataset_name: Name for the dataset
        :type dataset_name: str
        :param dataset_format: Format for the dataset
        :type dataset_format: str
        :param params: set of parameters to pass to URL
        :type params: dict
        """
        self.dataset_name = dataset_name
        if self.VALID_DATASET_FORMATS:
            if dataset_format and dataset_format not in self.VALID_DATASET_FORMATS:
                raise ValueError(
                    "Dataset format %s is not valid for %s"
                    % (dataset_format, dataset_name)
                )
        self.dataset_format = dataset_format
        self.params = params

    @property
    def dataset(self):
        """
        Qualify the dataset_name with the dataset_format if supplied
        :return: dataset name
        :rtype: str
        """
        if self.dataset_format:
            return ".".join([self.dataset_name, self.dataset_format])
        return self.dataset_name

    def url_path(self):
        """
        Get the correct URL Path for the Dataset
        :return:
        """
        args = ["datasets", self.dataset]
        return make_url(*args, **self.params)



# -*- coding: utf-8 -*-

__author__ = "isparks"

import unittest

from rwslib.builders import *
from ..common import obj_to_doc
from datetime import datetime


class TestClinicalData(unittest.TestCase):
    """Test ClinicalData classes"""

    def setUp(self):
        self.tested = ClinicalData("STUDY1", "DEV")(
            SubjectData("SITE1", "SUBJECT1")(
                StudyEventData("VISIT_1")(
                    FormData("TESTFORM_A")(
                        ItemGroupData()(
                            ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")
                        ),
                        ItemGroupData(item_group_repeat_key=2)(
                            ItemData("Field3", "ValueC")
                        ),
                    )
                )
            )
        )

    def test_basic(self):
        """Test there are 3 children"""
        self.assertEqual("STUDY1", self.tested.projectname)
        self.assertEqual("DEV", self.tested.environment)
        # Test default MetadataVersionOID
        self.assertEqual("1", self.tested.metadata_version_oid)

    def test_metadata_version_oid(self):
        """
        Test we can handle a MDV as a String
        """
        self.tested.metadata_version_oid = "2"
        doc = obj_to_doc(self.tested)
        self.assertEqual(
            doc.attrib["MetaDataVersionOID"], self.tested.metadata_version_oid
        )

    def test_metadata_version_oid_as_int(self):
        """
        Test that we can handle a MDV as an integer (which we are mandating in the IG)
        """
        self.tested.metadata_version_oid = 56
        doc = obj_to_doc(self.tested)
        self.assertEqual(
            doc.attrib["MetaDataVersionOID"], str(self.tested.metadata_version_oid)
        )

    def test_only_accepts_subjectdata(self):
        """Test that only SubjectData can be inserted"""
        tested = ClinicalData("STUDY1", "DEV")

        def do():
            tested << object()

        self.assertRaises(ValueError, do)

    def test_builder(self):
        """XML produced"""
        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.tag, "ClinicalData")

    def test_add_to_odm(self):
        """We can add multiple ClinicalData to an ODM"""
        odm = ODM("Some test case")
        odm << ClinicalData("Study1", "Dev")
        odm << ClinicalData("Study1", "Dev")
        tested = obj_to_doc(odm)
        self.assertEqual("ODM", tested.tag)
        self.assertTrue(2, len(list(tested)))

    def test_nix_prod_designation(self):
        """Prod gets removed"""
        odm = ODM("Some test case")
        odm << ClinicalData("Study1", "Prod")
        tested = obj_to_doc(odm)
        self.assertEqual("ODM", tested.tag)
        self.assertTrue(1, len(list(tested)))
        self.assertEqual("Study1", tested[0].attrib["StudyOID"])


class TestSubjectData(unittest.TestCase):
    """Test SubjectData classes"""

    def setUp(self):
        self.tested = SubjectData("SITE1", "SUBJECT1")(
            StudyEventData("VISIT_1")(
                FormData("TESTFORM_A")(
                    ItemGroupData()(
                        ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")
                    ),
                    ItemGroupData(item_group_repeat_key=2)(
                        ItemData("Field3", "ValueC")
                    ),
                )
            )
        )

    def test_basic(self):
        """Test there are 3 children"""
        self.assertEqual("SITE1", self.tested.sitelocationoid)
        self.assertEqual("SUBJECT1", self.tested.subject_key)
        # Default transaction type
        self.assertEqual("Update", self.tested.transaction_type)

    def test_invalid_transaction_type_direct_assign(self):
        """Test transaction type will not allow you to set to invalid choice"""

        def do():
            self.tested.transaction_type = "invalid"

        self.assertRaises(AttributeError, do)

    def test_children(self):
        """Test there is 1 child"""
        self.assertEqual(1, len(self.tested.study_events))

    def test_builder(self):
        """XML produced"""
        doc = obj_to_doc(self.tested)
        # Test default transaction tyoe
        self.assertEqual(doc.attrib["TransactionType"], "Update")
        self.assertEqual(doc.tag, "SubjectData")

    def test_only_add_studyeventdata_once(self):
        """Test that a StudyEventData object can only be added once"""
        sed = StudyEventData("V1")
        self.tested << sed

        def do():
            self.tested << sed

        self.assertRaises(ValueError, do)

    def test_does_not_accept_all_elements(self):
        """Test that,for example, ItemData cannot be accepted"""

        def do():
            self.tested << ItemData("Field1", "ValueC")

        self.assertRaises(ValueError, do)

    def test_accepts_auditrecord(self):
        """Test that AuditRecord can be inserted"""
        ar = AuditRecord(
            used_imputation_method=False, identifier="ABC1", include_file_oid=False
        )(
            UserRef("test_user"),
            LocationRef("test_site"),
            ReasonForChange("Testing"),
            DateTimeStamp(datetime.now()),
        )
        self.tested << ar
        self.assertEqual(self.tested.audit_record, ar)
        t = obj_to_doc(self.tested)
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(
            len(list(t)) == 3
        )  # 1 StudyEventData + 1 SiteRef + 1 AuditRecord

    def test_add_annotations(self):
        """Test we can add one or more annotations"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        for i in range(0, 4):
            self.tested << Annotation(
                comment=Comment("Some Comment %s" % i), flags=flags
            )
        t = obj_to_doc(self.tested)
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(
            len(list(t)) == 6
        )  # 1 StudyEventData + 1 SiteRef + 4 annotations

    def test_add_signature(self):
        """Test we can add one signature"""
        self.tested << Signature(
            signature_id="Some ID",
            user_ref=UserRef(oid="AUser"),
            location_ref=LocationRef(oid="ALocation"),
            signature_ref=SignatureRef(oid="ASignature"),
            date_time_stamp=DateTimeStamp(
                date_time=datetime(
                    year=2016, month=12, day=25, hour=12, minute=0, second=0
                )
            ),
        )
        t = obj_to_doc(self.tested)
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 3)  # 1 studyeventdata + 1 SiteRef + 1 signature

    def test_multiple_subject_data(self):
        """We can add multiple SubjectData to the Clinical Data"""
        cd = ClinicalData("Mediflex", "Prod")
        cd << SubjectData("Site1", "Subject1")
        cd << SubjectData("Site1", "Subject2")
        doc = obj_to_doc(cd)
        self.assertEqual(2, len(doc))

    def test_multiple_subject_data_location_oid(self):
        """Specify the LocationOID Type"""
        cd = ClinicalData("Mediflex", "Prod")
        cd << SubjectData("Site1", "Subject1",
                          location_oid_type=LocationOIDType.StudyEnvSiteNumber)
        doc = obj_to_doc(cd)
        site_ref = doc.find(".//SiteRef", namespaces={"odm": "http://www.cdisc.org/ns/odm/v1.3"})
        self.assertEqual('StudyEnvSiteNumber', site_ref.attrib["mdsol:LocationOIDType"])


class TestStudyEventData(unittest.TestCase):
    """Test StudyEventData classes"""

    def setUp(self):
        self.tested = StudyEventData("VISIT_1")(
            FormData("TESTFORM_A")(
                ItemGroupData()(
                    ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")
                ),
                ItemGroupData(item_group_repeat_key=2)(ItemData("Field3", "ValueC")),
            )
        )

    def test_transaction_type(self):
        """Test transaction type inserted if set"""
        self.tested.transaction_type = "Update"
        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.attrib["TransactionType"], self.tested.transaction_type)

    def test_builders_basic(self):
        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.attrib["StudyEventOID"], "VISIT_1")
        self.assertIsNone(doc.attrib.get("StudyEventRepeatKey"))
        self.assertEqual(len(doc), 1)
        self.assertEqual(doc.tag, "StudyEventData")

    def test_study_event_repeat_key(self):
        """ If supplied we export the study event repeat key"""
        tested = StudyEventData("VISIT_1", study_event_repeat_key="1")(
            FormData("TESTFORM_A")(
                ItemGroupData()(
                    ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")
                ),
                ItemGroupData(item_group_repeat_key=2)(ItemData("Field3", "ValueC")),
            )
        )
        t = obj_to_doc(tested)
        self.assertEqual("StudyEventData", t.tag)
        self.assertEqual("1", t.attrib["StudyEventRepeatKey"])

    def test_study_event_repeat_key_as_int(self):
        """ If supplied we export the study event repeat key"""
        tested = StudyEventData("VISIT_1", study_event_repeat_key=1)(
            FormData("TESTFORM_A")(
                ItemGroupData()(
                    ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")
                ),
                ItemGroupData(item_group_repeat_key=2)(ItemData("Field3", "ValueC")),
            )
        )
        t = obj_to_doc(tested)
        self.assertEqual("StudyEventData", t.tag)
        self.assertEqual("1", t.attrib["StudyEventRepeatKey"])

    def test_only_add_formdata_once(self):
        """Test that an FormData object can only be added once"""
        fd = FormData("FORM1")
        self.tested << fd

        def do():
            self.tested << fd

        self.assertRaises(ValueError, do)

    def test_add_annotations(self):
        """Test we can add one or more annotations"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        for i in range(0, 4):
            self.tested << Annotation(
                comment=Comment("Some Comment %s" % i), flags=flags
            )
        t = obj_to_doc(self.tested)
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 5)  # one formdata + 4 annotations

    def test_add_signature(self):
        """Test we can add one signature"""
        self.tested << Signature(
            signature_id="Some ID",
            user_ref=UserRef(oid="AUser"),
            location_ref=LocationRef(oid="ALocation"),
            signature_ref=SignatureRef(oid="ASignature"),
            date_time_stamp=DateTimeStamp(
                date_time=datetime(
                    year=2016, month=12, day=25, hour=12, minute=0, second=0
                )
            ),
        )
        t = obj_to_doc(self.tested)
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 2)  # 1 formdata + 1 signature

    def test_invalid_transaction_type_direct_assign(self):
        """Test transaction type will not allow you to set to invalid choice"""

        def do():
            self.tested.transaction_type = "invalid"

        self.assertRaises(AttributeError, do)

    def test_only_accepts_formdata(self):
        """Test that only FormData can be inserted"""

        def do():
            # Bzzzt. Should be ItemGroupData
            self.tested << ItemData("Field1", "ValueC")

        self.assertRaises(ValueError, do)


class TestFormData(unittest.TestCase):
    """Test FormData classes"""

    def setUp(self):
        self.tested = FormData("TESTFORM_A")(
            ItemGroupData()(ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")),
            ItemGroupData()(ItemData("Field3", "ValueC")),
            ItemGroupData()(ItemData("Field4", "ValueD")),
        )

    def test_children(self):
        """Test there are 3 children"""
        self.assertEqual(3, len(self.tested.itemgroups))

    def test_invalid_transaction_type(self):
        """Test transaction type will not allow you to set to invalid choice"""

        def do():
            FormData("MYFORM", transaction_type="invalid")

        self.assertRaises(AttributeError, do)

    def test_only_accepts_itemgroupdata(self):
        """Test that only ItemGroupData can be inserted"""

        def do():
            # Bzzzt. Should be ItemGroupData
            self.tested << ItemData("Field1", "ValueC")

        self.assertRaises(ValueError, do)

    def test_only_add_itemgroup_once(self):
        """Test that an ItemGroupData can only be added once"""
        igd = ItemGroupData()
        self.tested << igd

        def do():
            self.tested << igd

        self.assertRaises(ValueError, do)

    def test_builders_basic(self):
        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.attrib["FormOID"], "TESTFORM_A")
        self.assertEqual(len(doc), 3)
        self.assertEqual(doc.tag, "FormData")

    def test_transaction_type(self):
        """Test transaction type inserted if set"""
        self.tested.transaction_type = "Update"
        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.attrib["TransactionType"], self.tested.transaction_type)

    def test_invalid_transaction_type_direct_assign(self):
        """Test transaction type will not allow you to set to invalid choice"""

        def do():
            self.tested.transaction_type = "invalid"

        self.assertRaises(AttributeError, do)

    def test_form_repeat_key(self):
        """Test transaction type inserted if set"""
        tested = FormData("TESTFORM_A", form_repeat_key=9)(
            ItemGroupData()(ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB"))
        )
        doc = obj_to_doc(tested)
        self.assertEqual(doc.attrib["FormRepeatKey"], "9")

    def test_add_annotations(self):
        """Test we can add one or more annotations"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        for i in range(0, 4):
            self.tested << Annotation(
                comment=Comment("Some Comment %s" % i), flags=flags
            )
        t = obj_to_doc(self.tested)
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 7)  # three igdata + 4 annotations

    def test_add_signature(self):
        """Test we can add one signature"""
        self.tested << Signature(
            signature_id="Some ID",
            user_ref=UserRef(oid="AUser"),
            location_ref=LocationRef(oid="ALocation"),
            signature_ref=SignatureRef(oid="ASignature"),
            date_time_stamp=DateTimeStamp(
                date_time=datetime(
                    year=2016, month=12, day=25, hour=12, minute=0, second=0
                )
            ),
        )
        t = obj_to_doc(self.tested)
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 4)  # three igdata + 1 signature

    def test_lab_settings(self):
        tested = FormData("TESTFORM_A", lab_reference="A Lab", lab_type=LabType.Local)
        doc = obj_to_doc(tested)
        self.assertEqual(doc.attrib["mdsol:LaboratoryRef"], "A Lab")
        self.assertEqual(doc.attrib["mdsol:LaboratoryType"], "Local")

class TestItemGroupData(unittest.TestCase):
    """Test ItemGroupData classes"""

    def setUp(self):
        self.tested = ItemGroupData()(
            ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")
        )

    def test_children(self):
        """Test there are 2 children"""
        self.assertEqual(2, len(self.tested.items))

    def test_two_same_invalid(self):
        """Test adding a duplicate field causes error"""

        def do():
            self.tested << ItemData("Field1", "ValueC")

        self.assertRaises(ValueError, do)

    def test_only_accepts_itemdata(self):
        """Test that an ItemGroupData will only accept an ItemData element"""

        def do():
            self.tested << {"Field1": "ValueC"}

        self.assertRaises(ValueError, do)

    def test_invalid_transaction_type(self):
        def do():
            ItemGroupData(transaction_type="invalid")

        self.assertRaises(AttributeError, do)

    def test_builders_basic(self):
        doc = obj_to_doc(self.tested, "TESTFORM")
        self.assertEqual(doc.attrib["ItemGroupOID"], "TESTFORM")
        self.assertEqual(len(doc), 2)
        self.assertEqual(doc.tag, "ItemGroupData")

    def test_transaction_type(self):
        """Test transaction type inserted if set"""
        self.tested.transaction_type = "Context"
        doc = obj_to_doc(self.tested, "TESTFORM")
        self.assertEqual(doc.attrib["TransactionType"], "Context")

    def test_whole_item_group(self):
        """mdsol:Submission should be wholeitemgroup or SpecifiedItemsOnly"""
        doc = obj_to_doc(self.tested, "TESTFORM")
        self.assertEqual(doc.attrib["mdsol:Submission"], "SpecifiedItemsOnly")

        self.tested.whole_item_group = True
        doc = obj_to_doc(self.tested, "TESTFORM")
        self.assertEqual(doc.attrib["mdsol:Submission"], "WholeItemGroup")

    def test_add_annotations(self):
        """Test we can add one or more annotations"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        for i in range(0, 4):
            self.tested << Annotation(
                comment=Comment("Some Comment %s" % i), flags=flags
            )
        t = obj_to_doc(self.tested, "TESTFORM")
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 6)  # two itemdata + 4 annotations

    def test_add_annotations_on_create_multiple(self):
        """Test we can add one or more annotations at initialisation"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        annotations = [
            Annotation(comment=Comment("Some Comment %s" % i), flags=flags)
            for i in range(0, 4)
        ]
        # add a list of annotations
        igd = ItemGroupData(annotations=annotations)(
            ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")
        )
        t = obj_to_doc(igd, "TESTFORM")
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 6)  # two itemdata + 4 annotations

    def test_add_annotations_on_create_single(self):
        """Test we can add one or more annotations at initialisation with one"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        annotations = [
            Annotation(comment=Comment("Some Comment %s" % i), flags=flags)
            for i in range(0, 4)
        ]
        # add a list of annotations
        igd = ItemGroupData(annotations=annotations[0])(
            ItemData("Field1", "ValueA"), ItemData("Field2", "ValueB")
        )
        t = obj_to_doc(igd, "TESTFORM")
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 3)  # two itemdata + 4 annotations

    def test_add_signature(self):
        """Test we can add one signature"""
        self.tested << Signature(
            signature_id="Some ID",
            user_ref=UserRef(oid="AUser"),
            location_ref=LocationRef(oid="ALocation"),
            signature_ref=SignatureRef(oid="ASignature"),
            date_time_stamp=DateTimeStamp(
                date_time=datetime(
                    year=2016, month=12, day=25, hour=12, minute=0, second=0
                )
            ),
        )
        t = obj_to_doc(self.tested, "TESTFORM")
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 3)  # two itemdata + 1 signature


class TestItemData(unittest.TestCase):
    """Test ItemData classes"""

    def setUp(self):
        self.tested = ItemData("FIELDA", "TEST")

    def test_unset(self):
        """Test unset value"""
        itmdata = ItemData("FIELD")
        doc = obj_to_doc(itmdata)
        self.assertRaises(KeyError, doc.attrib.__getitem__, "Value")
        self.assertRaises(KeyError, doc.attrib.__getitem__, "IsNull")

    def test_basic(self):
        tested = self.tested
        self.assertEqual(tested.itemoid, "FIELDA")
        self.assertEqual(tested.value, "TEST")
        self.assertEqual(tested.lock, None)
        self.assertEqual(tested.freeze, None)
        self.assertEqual(tested.verify, None)

    def test_only_accepts_itemdata(self):
        """Test that an ItemData will not accept any old object"""
        with self.assertRaises(ValueError):
            self.tested << {"Field1": "ValueC"}

    def test_accepts_query(self):
        """Test that an ItemData will accept a query"""
        query = MdsolQuery()
        self.tested << query
        self.assertEqual(query, self.tested.queries[0])

    def test_accepts_measurement_unit_ref(self):
        """Test that an ItemData will accept a measurement unit ref"""
        mur = MeasurementUnitRef("Celsius")
        self.tested << mur
        self.assertEqual(mur, self.tested.measurement_unit_ref)

    def test_isnull_not_set(self):
        """Isnull should not be set where we have a value not in '', None"""
        doc = obj_to_doc(self.tested)

        # Check IsNull attribute is missing
        def do():
            doc.attrib["IsNull"]

        self.assertRaises(KeyError, do)

    def test_specify(self):
        """Test specify"""
        specify_value = "A Specify"
        self.tested.specify_value = specify_value
        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.attrib["mdsol:SpecifyValue"], specify_value)

    def test_freeze_lock_verify(self):
        tested = ItemData("FIELDA", "TEST", lock=True, verify=True, freeze=False)
        self.assertEqual(tested.lock, True)
        self.assertEqual(tested.freeze, False)
        self.assertEqual(tested.verify, True)

    def test_builder(self):
        """Test building XML"""
        tested = ItemData("FIELDA", "TEST", lock=True, verify=True, freeze=False)

        tested << AuditRecord(
            edit_point=AuditRecord.EDIT_DATA_MANAGEMENT,
            used_imputation_method=False,
            identifier="x2011",
            include_file_oid=False,
        )(
            UserRef("Fred"),
            LocationRef("Site102"),
            ReasonForChange("Data Entry Error"),
            DateTimeStamp(datetime(2015, 9, 11, 10, 15, 22, 80)),
        )
        tested << MdsolQuery()
        tested << MeasurementUnitRef("Celsius")

        doc = obj_to_doc(tested)

        self.assertEqual(doc.attrib["ItemOID"], "FIELDA")
        self.assertEqual(doc.attrib["Value"], "TEST")
        self.assertEqual(doc.attrib["mdsol:Verify"], "Yes")
        self.assertEqual(doc.attrib["mdsol:Lock"], "Yes")
        self.assertEqual(doc.attrib["mdsol:Freeze"], "No")
        self.assertEqual(doc.tag, "ItemData")
        self.assertEqual("AuditRecord", list(doc)[0].tag)
        self.assertEqual("MeasurementUnitRef", list(doc)[1].tag)
        self.assertEqual("mdsol:Query", list(doc)[2].tag)

    def test_transaction_type(self):
        tested = self.tested
        tested.transaction_type = "Update"
        doc = obj_to_doc(tested)
        self.assertEqual(doc.attrib["TransactionType"], "Update")

    def test_null_value(self):
        """Null or empty string values are treated specially with IsNull property and no value"""
        tested = self.tested
        tested.value = ""
        doc = obj_to_doc(tested)
        self.assertEqual(doc.attrib["IsNull"], "Yes")

        # Check Value attribute is also missing
        def do():
            doc.attrib["Value"]

        self.assertRaises(KeyError, do)

    def test_invalid_transaction_type(self):
        def do():
            ItemData("A", "val", transaction_type="invalid")

        self.assertRaises(AttributeError, do)

    def test_add_annotations(self):
        """Test we can add one or more annotations"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        for i in range(0, 4):
            self.tested << Annotation(
                comment=Comment("Some Comment %s" % i), flags=flags
            )
        t = obj_to_doc(self.tested)
        self.assertEqual(self.__class__.__name__[4:], t.tag)
        self.assertTrue(len(list(t)) == 4)  # one formdata + 4 annotations


class TestUserRef(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            UserRef("Gertrude") << object()

    def test_builder(self):
        """Test building XML"""
        tested = UserRef("Fred")
        doc = obj_to_doc(tested)

        self.assertEqual(doc.attrib["UserOID"], "Fred")
        self.assertEqual(doc.tag, "UserRef")


class TestLocationRef(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            LocationRef("Nowhereville") << object()

    def test_builder(self):
        """Test building XML"""
        tested = LocationRef("Gainesville")
        doc = obj_to_doc(tested)

        self.assertEqual(doc.attrib["LocationOID"], "Gainesville")
        self.assertEqual(doc.tag, "LocationRef")

    def test_builder_int_oid(self):
        """Test building XML"""
        tested = LocationRef(12)
        doc = obj_to_doc(tested)

        self.assertEqual(doc.attrib["LocationOID"], "12")
        self.assertEqual(doc.tag, "LocationRef")


class TestReasonForChange(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            ReasonForChange("Because I wanted to") << object()

    def test_builder(self):
        """Test building XML"""
        tested = ReasonForChange("Testing 1..2..3")
        doc = obj_to_doc(tested)

        self.assertEqual("Testing 1..2..3", doc.text)
        self.assertEqual(doc.tag, "ReasonForChange")


class TestDateTimeStamp(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            DateTimeStamp(datetime.now()) << object()

    def test_builder_with_datetime(self):
        dt = datetime(2015, 9, 11, 10, 15, 22, 80)
        tested = DateTimeStamp(dt)
        doc = obj_to_doc(tested)

        self.assertEqual(dt_to_iso8601(dt), doc.text)
        self.assertEqual(doc.tag, "DateTimeStamp")

    def test_builder_with_string(self):
        dt = "2009-02-04T14:10:32-05:00"
        tested = DateTimeStamp(dt)
        doc = obj_to_doc(tested)
        self.assertEqual(dt, doc.text)
        self.assertEqual(doc.tag, "DateTimeStamp")


class TestAuditRecord(unittest.TestCase):
    def setUp(self):
        self.tested = AuditRecord(
            edit_point=AuditRecord.EDIT_DATA_MANAGEMENT,
            used_imputation_method=False,
            identifier="X2011",
            include_file_oid=False,
        )
        self.tested << UserRef("Fred")
        self.tested << LocationRef("Site102")
        self.tested << ReasonForChange("Data Entry Error")
        self.tested << DateTimeStamp(datetime(2015, 9, 11, 10, 15, 22, 80))

    def test_identifier_must_not_start_digit(self):
        with self.assertRaises(AttributeError):
            AuditRecord(identifier="2011")

        with self.assertRaises(AttributeError):
            AuditRecord(identifier="*Hello")

        # Underscore OK
        ar = AuditRecord(identifier="_Hello")
        self.assertEqual("_Hello", ar.audit_id)

        # Letter OK
        ar = AuditRecord(identifier="Hello")
        self.assertEqual("Hello", ar.audit_id)

    def test_accepts_no_invalid_children(self):
        with self.assertRaises(ValueError):
            AuditRecord() << object()

    def test_invalid_edit_point(self):
        with self.assertRaises(AttributeError):
            AuditRecord(edit_point="Blah")

    def test_builder(self):
        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.tag, "AuditRecord")
        self.assertEqual(AuditRecord.EDIT_DATA_MANAGEMENT, doc.attrib["EditPoint"])
        self.assertEqual("No", doc.attrib["UsedImputationMethod"])
        self.assertEqual("No", doc.attrib["mdsol:IncludeFileOID"])
        self.assertEqual("UserRef", list(doc)[0].tag)
        self.assertEqual("LocationRef", list(doc)[1].tag)
        self.assertEqual("DateTimeStamp", list(doc)[2].tag)
        self.assertEqual("ReasonForChange", list(doc)[3].tag)

    def test_no_user_ref(self):
        """Test with no user ref should fail on build with a ValueError"""
        self.tested.user_ref = None
        with self.assertRaises(ValueError) as err:
            doc = obj_to_doc(self.tested)
            self.assertIn("UserRef", err.exception.message)

    def test_no_location_ref(self):
        """Test with no location ref should fail on build with a ValueError"""
        self.tested.location_ref = None
        with self.assertRaises(ValueError) as err:
            doc = obj_to_doc(self.tested)
            self.assertIn("LocationRef", err.exception.message)

    def test_no_datetime_stamp(self):
        """Test with no datetimestamp should fail on build with a ValueError"""
        self.tested.date_time_stamp = None
        with self.assertRaises(ValueError) as err:
            doc = obj_to_doc(self.tested)
            self.assertIn("DateTimeStamp", err.exception.message)


class TestSignatureRef(unittest.TestCase):
    def test_creates_expected_element(self):
        """We get the Signature Ref element"""
        t = SignatureRef("ASIGNATURE")
        doc = obj_to_doc(t)
        self.assertEqual("SignatureRef", doc.tag)
        self.assertEqual("ASIGNATURE", doc.attrib["SignatureOID"])


class TestSignature(unittest.TestCase):
    def test_creates_expected_element(self):
        """We create a Signature element"""
        t = Signature(
            signature_id="Some ID",
            user_ref=UserRef(oid="AUser"),
            location_ref=LocationRef(oid="ALocation"),
            signature_ref=SignatureRef(oid="ASignature"),
            date_time_stamp=DateTimeStamp(
                date_time=datetime(
                    year=2016, month=12, day=25, hour=12, minute=0, second=0
                )
            ),
        )
        doc = obj_to_doc(t)
        self.assertEqual("Signature", doc.tag)
        self.assertEqual("Some ID", doc.attrib["ID"])
        # all four elements are present
        self.assertTrue(len(list(doc)) == 4)

    def test_creates_expected_element_no_id(self):
        """We create a Signature element without an ID"""
        t = Signature(
            user_ref=UserRef(oid="AUser"),
            location_ref=LocationRef(oid="ALocation"),
            signature_ref=SignatureRef(oid="ASignature"),
            date_time_stamp=DateTimeStamp(
                date_time=datetime(
                    year=2016, month=12, day=25, hour=12, minute=0, second=0
                )
            ),
        )
        doc = obj_to_doc(t)
        self.assertEqual("Signature", doc.tag)
        self.assertTrue("ID" not in doc.attrib)
        # all four elements are present
        self.assertTrue(len(list(doc)) == 4)

    def test_all_elements_are_required(self):
        """All the sub-elements are required"""
        all = dict(
            user_ref=UserRef(oid="AUser"),
            location_ref=LocationRef(oid="ALocation"),
            signature_ref=SignatureRef(oid="ASignature"),
            date_time_stamp=DateTimeStamp(
                date_time=datetime(
                    year=2016, month=12, day=25, hour=12, minute=0, second=0
                )
            ),
        )
        t0 = Signature()
        with self.assertRaises(ValueError) as exc:
            doc = obj_to_doc(t0)
        self.assertEqual("User Reference not set.", str(exc.exception))
        t1 = Signature(user_ref=all.get("user_ref"))
        with self.assertRaises(ValueError) as exc:
            doc = obj_to_doc(t1)
        self.assertEqual("Location Reference not set.", str(exc.exception))
        t2 = Signature(
            user_ref=all.get("user_ref"), location_ref=all.get("location_ref")
        )
        with self.assertRaises(ValueError) as exc:
            doc = obj_to_doc(t2)
        self.assertEqual("Signature Reference not set.", str(exc.exception))
        t3 = Signature(
            user_ref=all.get("user_ref"),
            location_ref=all.get("location_ref"),
            signature_ref=all.get("signature_ref"),
        )
        with self.assertRaises(ValueError) as exc:
            doc = obj_to_doc(t3)
        self.assertEqual("DateTime not set.", str(exc.exception))

    def test_signature_builder(self):
        """"""
        tested = Signature(signature_id="Some ID")
        all = dict(
            user_ref=UserRef(oid="AUser"),
            location_ref=LocationRef(oid="ALocation"),
            signature_ref=SignatureRef(oid="ASignature"),
            date_time_stamp=DateTimeStamp(
                date_time=datetime(
                    year=2016, month=12, day=25, hour=12, minute=0, second=0
                )
            ),
        )
        for child in all.values():
            tested << child
        doc = obj_to_doc(tested)
        self.assertEqual("Signature", doc.tag)
        self.assertEqual("Some ID", doc.attrib["ID"])
        # all four elements are present
        self.assertTrue(len(list(doc)) == 4)

    def test_signature_builder_with_invalid_input(self):
        """"""
        tested = Signature(signature_id="Some ID")
        with self.assertRaises(ValueError) as exc:
            tested << ItemData(itemoid="GENDER", value="MALE")
        self.assertEqual(
            "Signature cannot accept a child element of type ItemData",
            str(exc.exception),
        )


class TestAnnotation(unittest.TestCase):
    """ Test Annotation classes """

    def test_happy_path(self):
        """ Simple Annotation with a single flag and comment"""
        tested = Annotation(annotation_id="APPLE", seqnum=1)
        f = Flag(
            flag_value=FlagValue("Some value", codelist_oid="ANOID"),
            flag_type=FlagType("Some type", codelist_oid="ANOTHEROID"),
        )
        c = Comment("Some Comment")
        tested << f
        tested << c
        t = obj_to_doc(tested)
        self.assertEqual("Annotation", t.tag)
        self.assertEqual("1", t.attrib["SeqNum"])
        self.assertEqual("APPLE", t.attrib["ID"])
        self.assertTrue(len(list(t)) == 2)

    def test_happy_path_id_optional(self):
        """ Simple Annotation with a single flag and comment, no ID"""
        tested = Annotation(seqnum=1)
        f = Flag(
            flag_value=FlagValue("Some value", codelist_oid="ANOID"),
            flag_type=FlagType("Some type", codelist_oid="ANOTHEROID"),
        )
        c = Comment("Some Comment")
        tested << f
        tested << c
        t = obj_to_doc(tested)
        self.assertEqual("Annotation", t.tag)
        self.assertEqual("1", t.attrib["SeqNum"])
        self.assertNotIn("ID", t.attrib)
        self.assertTrue(len(list(t)) == 2)

    def test_happy_path_seqnum_defaulted(self):
        """ Simple Annotation with a single flag and comment, SeqNum missing"""
        tested = Annotation()
        f = Flag(
            flag_value=FlagValue("Some value", codelist_oid="ANOID"),
            flag_type=FlagType("Some type", codelist_oid="ANOTHEROID"),
        )
        c = Comment("Some Comment")
        tested << f
        tested << c
        t = obj_to_doc(tested)
        self.assertEqual("Annotation", t.tag)
        self.assertEqual("1", t.attrib["SeqNum"])
        self.assertTrue(len(list(t)) == 2)

    def test_happy_path_multiple_flags(self):
        """ Simple Annotation with a multiple flags and comment"""
        tested = Annotation()
        c = Comment("Some Comment")
        # Add some flags
        for i in range(0, 3):
            tested << Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
        tested << c
        t = obj_to_doc(tested)
        self.assertEqual("Annotation", t.tag)
        self.assertTrue(len(list(t)) == 4)

    def test_happy_path_multiple_flags_on_init(self):
        """ Simple Annotation with a multiple flags and comment created at init"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        tested = Annotation(comment=Comment("Some Comment"), flags=flags)
        t = obj_to_doc(tested)
        self.assertEqual("Annotation", t.tag)
        self.assertTrue(len(list(t)) == 4)

    def test_happy_path_flag_on_init(self):
        """ Simple Annotation with a single flag and comment created at init"""
        flags = [
            Flag(
                flag_value=FlagValue("Some value %s" % i, codelist_oid="ANOID%s" % i),
                flag_type=FlagType("Some type %s" % i, codelist_oid="ANOTHEROID%s" % i),
            )
            for i in range(0, 3)
        ]
        tested = Annotation(comment=Comment("Some Comment"), flags=flags[0])
        t = obj_to_doc(tested)
        self.assertEqual("Annotation", t.tag)
        self.assertTrue(len(list(t)) == 2)

    def test_not_flag_on_init(self):
        """ Simple Annotation with not a flag raises an exception and comment created at init"""
        notflags = ItemData(itemoid="GENDER", value="MALE")
        with self.assertRaises(AttributeError) as exc:
            tested = Annotation(comment=Comment("Some Comment"), flags=notflags)
        self.assertEqual(
            "Flags attribute should be an iterable or Flag", str(exc.exception)
        )

    def test_only_accept_valid_children(self):
        """ Annotation can only take one or more Flags and one Comment"""
        tested = Annotation(annotation_id="An Annotation")
        with self.assertRaises(ValueError) as exc:
            tested << ItemData(itemoid="GENDER", value="MALE")
        self.assertEqual(
            "Annotation cannot accept a child element of type ItemData",
            str(exc.exception),
        )
        tested << Comment("A comment")
        with self.assertRaises(ValueError) as exc:
            tested << Comment("Another Comment")
        self.assertEqual(
            "Annotation already has a Comment element set.", str(exc.exception)
        )

    def test_only_valid_id_accepted(self):
        """ Annotation ID must be a non empty string"""
        for nonsense in ("", "     "):
            with self.assertRaises(AttributeError) as exc:
                tested = Annotation(annotation_id=nonsense)
            self.assertEqual(
                "Invalid ID value supplied",
                str(exc.exception),
                "Value should raise with '%s'" % nonsense,
            )

    def test_only_valid_seqnum_accepted(self):
        """ Annotation ID must be a non empty string"""
        for nonsense in ("apple", "     ", -1):
            with self.assertRaises(AttributeError) as exc:
                tested = Annotation(seqnum=nonsense)
            self.assertEqual(
                "Invalid SeqNum value supplied",
                str(exc.exception),
                "Value should raise with '%s'" % nonsense,
            )

    def test_need_flags(self):
        """ Annotation needs a Flag """
        tested = Annotation(comment=Comment("A comment"))
        with self.assertRaises(ValueError) as exc:
            t = obj_to_doc(tested)
        self.assertEqual("Flag is not set.", str(exc.exception))

    def test_transaction_type(self):
        """ Annotation can take a transaction type """
        tested = Annotation(
            flags=Flag(
                flag_value=FlagValue("Some value", codelist_oid="ANOID"),
                flag_type=FlagType("Some type", codelist_oid="ANOTHEROID"),
            ),
            comment=Comment("A comment"),
            transaction_type="Update",
        )
        t = obj_to_doc(tested)
        self.assertEqual("Annotation", t.tag)
        self.assertEqual("Update", t.attrib["TransactionType"])


class TestAnnotations(unittest.TestCase):
    def test_happy_path(self):
        """We create a Annotations object and add annotations to it"""
        obj = Annotations()
        obj << Annotation(annotation_id="1")(
            Flag()(FlagValue("test 1", codelist_oid="MILESTONE"))
        )
        obj << Annotation(annotation_id="2")(
            Flag()(FlagValue("test 2", codelist_oid="MILESTONE"))
        )
        obj << Annotation(annotation_id="3")(
            Flag()(FlagValue("test 3", codelist_oid="MILESTONE"))
        )
        tested = obj_to_doc(obj)
        self.assertEqual("Annotations", tested.tag)
        self.assertEqual(3, len(list(tested)))

    def test_sad_path(self):
        """We create a Annotations object and can't add a flag"""
        obj = Annotations()
        with self.assertRaises(ValueError) as exc:
            obj << Flag()(FlagValue("test 1", codelist_oid="MILESTONE"))
        self.assertEqual(
            "Annotations cannot accept a child element of type Flag", str(exc.exception)
        )


class TestFlag(unittest.TestCase):
    """ Test Flag classes """

    def test_happy_path(self):
        """Create a Flag object"""
        tested = Flag()
        tested << FlagValue("Some value", codelist_oid="ANOID")
        tested << FlagType("Some type", codelist_oid="ANOTHEROID")
        t = obj_to_doc(tested)
        self.assertEqual("Flag", t.tag)
        self.assertTrue(len(list(t)) == 2)

    def test_no_value(self):
        """No FlagValue is an exception"""
        tested = Flag()
        tested << FlagType("Some type", codelist_oid="ANOTHEROID")
        with self.assertRaises(ValueError) as exc:
            t = obj_to_doc(tested)
        self.assertEqual("FlagValue is not set.", str(exc.exception))

    def test_only_expected_types(self):
        """We can only add Flag-type elements"""
        tested = Flag()
        with self.assertRaises(ValueError) as exc:
            tested << ItemData(itemoid="GENDER", value="MALE")
        self.assertEqual(
            "Flag cannot accept a child element of type ItemData", str(exc.exception)
        )

    def test_only_expected_types_instance_vars(self):
        """We can only add Flag-type elements"""
        with self.assertRaises(ValueError) as exc:
            tested = Flag(flag_type=ItemData(itemoid="GENDER", value="MALE"))
        self.assertEqual(
            "Flag cannot accept a child element of type ItemData", str(exc.exception)
        )
        with self.assertRaises(ValueError) as exc:
            tested = Flag(flag_value=ItemData(itemoid="GENDER", value="MALE"))
        self.assertEqual(
            "Flag cannot accept a child element of type ItemData", str(exc.exception)
        )


class TestFlagType(unittest.TestCase):
    """ Test FlagType classes """

    def test_happy_path(self):
        """Create a FlagType object"""
        tested = FlagType("A Type")
        tested.codelist_oid = "ANOID"
        t = obj_to_doc(tested)
        self.assertEqual("FlagType", t.tag)
        self.assertEqual("ANOID", t.attrib["CodeListOID"])
        self.assertEqual("A Type", t.text)

    def test_no_oid_exception(self):
        """Create a FlagType object without a CodeListOID is an exception"""
        tested = FlagType("A Type")
        with self.assertRaises(ValueError) as exc:
            t = obj_to_doc(tested)
        self.assertEqual("CodeListOID not set.", str(exc.exception))

    def test_invalid_oid_exception(self):
        """Assigning a nonsensical value is an error"""
        tested = FlagType("A Type")
        for nonsense in (None, "", "   "):
            with self.assertRaises(AttributeError) as exc:
                tested.codelist_oid = nonsense
            self.assertEqual("Empty CodeListOID value is invalid.", str(exc.exception))

    def test_invalid_oid_exception_at_creation(self):
        """Assigning a nonsensical value is an error"""
        with self.assertRaises(AttributeError) as exc:
            tested = FlagType("A Type", codelist_oid="")
        self.assertEqual("Empty CodeListOID value is invalid.", str(exc.exception))


class TestFlagValue(unittest.TestCase):
    """ Test FlagValue classes """

    def test_happy_path(self):
        """Create a FlagValue object"""
        tested = FlagValue("A Value")
        tested.codelist_oid = "ANOID"
        t = obj_to_doc(tested)
        self.assertEqual("FlagValue", t.tag)
        self.assertEqual("ANOID", t.attrib["CodeListOID"])
        self.assertEqual("A Value", t.text)

    def test_no_oid_exception(self):
        """Create a FlagType object without a CodeListOID is an exception"""
        tested = FlagValue("A Type")
        with self.assertRaises(ValueError) as exc:
            t = obj_to_doc(tested)
        self.assertEqual("CodeListOID not set.", str(exc.exception))

    def test_invalid_oid_exception(self):
        """Assigning a nonsensical value is an error"""
        tested = FlagValue("A Type")
        for nonsense in (None, "", "   "):
            with self.assertRaises(AttributeError) as exc:
                tested.codelist_oid = nonsense
            self.assertEqual("Empty CodeListOID value is invalid.", str(exc.exception))

    def test_invalid_oid_exception_at_creation(self):
        """Assigning a nonsensical value is an error"""
        with self.assertRaises(AttributeError) as exc:
            tested = FlagValue("A Value", codelist_oid="")
        self.assertEqual("Empty CodeListOID value is invalid.", str(exc.exception))


class TestComment(unittest.TestCase):
    """ Test Comment classes """

    def test_happy_path(self):
        """Creating a valid Comment, no problems"""
        tested = Comment()
        tested.text = "Some comment"
        tested.sponsor_or_site = "Site"
        t = obj_to_doc(tested)
        self.assertEqual("Comment", t.tag)
        self.assertEqual("Site", t.attrib["SponsorOrSite"])
        self.assertEqual("Some comment", t.text)

    def test_happy_path_no_commenter(self):
        """Creating a valid Comment without a commenter, no problems"""
        tested = Comment()
        tested.text = "Some comment"
        t = obj_to_doc(tested)
        self.assertEqual("Comment", t.tag)
        self.assertNotIn("SponsorOrSite", t.attrib)
        self.assertEqual("Some comment", t.text)

    def test_invalid_commenter(self):
        """Creating a valid Comment with an invalid commenter, get an exception"""
        tested = Comment()
        tested.text = "Some comment"
        with self.assertRaises(AttributeError) as exc:
            tested.sponsor_or_site = "Some guy off the street"
        self.assertEqual(
            "Comment sponsor_or_site value of Some guy off the street is not valid",
            str(exc.exception),
        )

    def test_invalid_no_comment(self):
        """Creating a invalid Comment, get an exception"""
        tested = Comment()
        with self.assertRaises(ValueError) as exc:
            t = obj_to_doc(tested)
        self.assertEqual("Text is not set.", str(exc.exception))

    def test_invalid_text_comment(self):
        """Creating a Comment with invalid text, get an exception"""
        tested = Comment()
        for nonsense in (None, "", "    "):
            with self.assertRaises(AttributeError) as exc:
                tested.text = nonsense
            self.assertEqual("Empty text value is invalid.", str(exc.exception))


class TestSourceID(unittest.TestCase):
    def test_create_source_id(self):
        """We can create a source ID"""
        obj = SourceID("12345")
        tested = obj_to_doc(obj)
        self.assertEqual("SourceID", tested.tag)
        self.assertEqual("12345", tested.text)

    def test_add_to_audit(self):
        """We can add a SourceID to an Audit"""
        record = AuditRecord()
        record << UserRef("glow1")
        record << LocationRef("hillview")
        record << DateTimeStamp(get_utc_date().isoformat())
        record << SourceID("12345")
        tested = obj_to_doc(record)
        self.assertEqual("AuditRecord", tested.tag)
        self.assertEqual("SourceID", list(tested)[-1].tag)
        self.assertEqual("12345", list(tested)[-1].text)


class TestSiteRef(unittest.TestCase):
    def test_uuid_type(self):
        """We can define a SiteRef using a UUID"""
        siteref = SiteRef(oid="E20DEF2D-0CD4-4B3A-B963-AC7D592CB85B")
        siteref.add_attribute("LocationOIDType", "SiteUUID")
        tested = obj_to_doc(siteref)
        self.assertEqual("SiteRef", tested.tag)
        self.assertEqual(
            "E20DEF2D-0CD4-4B3A-B963-AC7D592CB85B", tested.get("LocationOID")
        )
        self.assertEqual("SiteUUID", tested.get("mdsol:LocationOIDType"))

    def test_uuid_type(self):
        """We can define a SiteRef using a UUID"""
        siteref = SiteRef(oid="E20DEF2D-0CD4-4B3A-B963-AC7D592CB85B")
        siteref.add_attribute("LocationOIDType", "SiteUUID")
        tested = obj_to_doc(siteref)
        self.assertEqual("SiteRef", tested.tag)
        self.assertEqual(
            "E20DEF2D-0CD4-4B3A-B963-AC7D592CB85B", tested.get("LocationOID")
        )
        self.assertEqual("SiteUUID", tested.get("mdsol:LocationOIDType"))

    def test_oid_type_init_site_uuid(self):
        siteref = SiteRef(oid="E20DEF2D-0CD4-4B3A-B963-AC7D592CB85B", location_oid_type=LocationOIDType.SiteUUID)
        tested = obj_to_doc(siteref)
        self.assertEqual("SiteRef", tested.tag)
        self.assertEqual(
            "E20DEF2D-0CD4-4B3A-B963-AC7D592CB85B", tested.get("LocationOID")
        )
        self.assertEqual("SiteUUID", tested.get("mdsol:LocationOIDType"))

    def test_oid_type_init_site_number(self):
        siteref = SiteRef(oid="001", location_oid_type=LocationOIDType.SiteNumber)
        tested = obj_to_doc(siteref)
        self.assertEqual("SiteRef", tested.tag)
        self.assertEqual(
            "001", tested.get("LocationOID")
        )
        self.assertEqual("SiteNumber", tested.get("mdsol:LocationOIDType"))

    def test_oid_type_init_study_env_site_number(self):
        siteref = SiteRef(oid="01", location_oid_type=LocationOIDType.StudyEnvSiteNumber)
        tested = obj_to_doc(siteref)
        self.assertEqual("SiteRef", tested.tag)
        self.assertEqual(
            "01", tested.get("LocationOID")
        )
        self.assertEqual("StudyEnvSiteNumber", tested.get("mdsol:LocationOIDType"))

    def test_oid_type_study_env_site_number(self):
        siteref = SiteRef(oid="01", study_env_site_number="001")
        tested = obj_to_doc(siteref)
        self.assertEqual("SiteRef", tested.tag)
        self.assertEqual(
            "01", tested.get("LocationOID")
        )
        self.assertEqual("001", tested.get("mdsol:StudyEnvSiteNumber"))



# -*- coding: utf-8 -*-
__author__ = "isparks"

import unittest
from rwslib.extras.audit_event import parser
import datetime


class TestUtils(unittest.TestCase):
    """Tests of utilities in the parser unit"""

    def test_make_int(self):
        self.assertEqual(-1, parser.make_int(None))
        self.assertEqual(-1, parser.make_int(""))
        self.assertEqual(5, parser.make_int("5"))
        self.assertEqual(6, parser.make_int(u"6"))

        with self.assertRaises(ValueError):
            self.assertEqual(-1, parser.make_int("five"))


class ParserTestCaseBase(unittest.TestCase):
    def setUp(self):
        class EventReporter(object):
            """Parser that will collect data"""

            def __init__(self):
                """We will just collect contexts as they fire from the default"""
                self.contexts = []
                self.subjects_created = 0

            def SubjectCreated(self, context):
                """As a test of calling non-default handler, we'll count subjects"""
                self.default(context)
                self.subjects_created += 1

            def default(self, context):
                self.contexts.append(context)

            @property
            def context(self):
                """Return the first context"""
                return self.contexts[0] if self.contexts else None

            def close(self):
                self.contexts = []
                return self.subjects_created

        self.count = 0
        self.eventer = EventReporter()

    @property
    def context(self):
        """Shortcut"""
        return self.eventer.context

    def parse(self, data):
        """Parse data, causing EventReporter to fire"""
        self.count = parser.parse(data, self.eventer)


class ParserTestCase(ParserTestCaseBase):
    def test_subject_created(self):
        """Test basics with subject created"""

        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MOVE-2014(DEV)" MetaDataVersionOID="2867" mdsol:AuditSubCategoryName="SubjectCreated">
    <SubjectData SubjectKey="538bdc4d-78b7-4ff9-a59c-3d13c8d8380b" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="01" TransactionType="Upsert">
      <AuditRecord>
        <UserRef UserOID="isparks" />
        <LocationRef LocationOID="1001" />
        <DateTimeStamp>2014-08-13T10:40:06</DateTimeStamp>
        <ReasonForChange />
        <SourceID>6434193</SourceID>
      </AuditRecord>
      <SiteRef LocationOID="1001" />
    </SubjectData>
  </ClinicalData></ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("SubjectCreated", sc.subcategory)
        self.assertEqual("MOVE-2014(DEV)", sc.study_oid)
        self.assertEqual(2867, sc.metadata_version)
        self.assertEqual("01", sc.subject.name)
        self.assertEqual("538bdc4d-78b7-4ff9-a59c-3d13c8d8380b", sc.subject.key)
        self.assertEqual(6434193, sc.audit_record.source_id)
        self.assertEqual(None, sc.audit_record.reason_for_change)
        self.assertEqual(
            datetime.datetime(2014, 8, 13, 10, 40, 6), sc.audit_record.datetimestamp
        )
        self.assertEqual("1001", sc.audit_record.location_oid)
        self.assertEqual("isparks", sc.audit_record.user_oid)

        # Check the SubjectCreated event fired
        self.assertEqual(1, self.eventer.subjects_created)

    def test_data_entered(self):
        """Test data entered with folder refs, reasons for change etc"""

        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
   <ClinicalData StudyOID="MOVE-2014(DEV)" MetaDataVersionOID="2867" mdsol:AuditSubCategoryName="EnteredWithChangeCode">
      <SubjectData SubjectKey="538bdc4d-78b7-4ff9-a59c-3d13c8d8380b" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="01">
         <SiteRef LocationOID="1001" />
         <StudyEventData StudyEventOID="VISIT1" StudyEventRepeatKey="VISIT1[1]" mdsol:InstanceId="227392">
            <FormData FormOID="VISIT" FormRepeatKey="1" mdsol:DataPageId="853098">
               <ItemGroupData ItemGroupOID="VISIT" mdsol:RecordId="1693434">
                  <ItemData ItemOID="VISIT.VISITDAT" TransactionType="Upsert" Value="7 Aug 2014">
                     <AuditRecord>
                        <UserRef UserOID="isparks" />
                        <LocationRef LocationOID="1001" />
                        <DateTimeStamp>2014-08-13T10:53:57</DateTimeStamp>
                        <ReasonForChange>Data Entry Error</ReasonForChange>
                        <SourceID>6434227</SourceID>
                     </AuditRecord>
                  </ItemData>
               </ItemGroupData>
            </FormData>
         </StudyEventData>
      </SubjectData>
   </ClinicalData>
</ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("EnteredWithChangeCode", sc.subcategory)
        self.assertEqual("VISIT1[1]", sc.event.repeat_key)
        self.assertEqual("Upsert", sc.item.transaction_type)
        self.assertEqual("VISIT", sc.form.oid)
        self.assertEqual(1, sc.form.repeat_key)
        self.assertEqual("VISIT", sc.itemgroup.oid)
        self.assertEqual("7 Aug 2014", sc.item.value)
        self.assertEqual("VISIT.VISITDAT", sc.item.oid)
        self.assertEqual("Data Entry Error", sc.audit_record.reason_for_change)
        self.assertEqual(227392, sc.event.instance_id)
        self.assertEqual(853098, sc.form.datapage_id)
        self.assertEqual(1693434, sc.itemgroup.record_id)

    def test_query(self):
        """Test data entered with queries"""
        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MOVE-2014(DEV)" MetaDataVersionOID="2867" mdsol:AuditSubCategoryName="QueryOpen">
    <SubjectData SubjectKey="a7d8d74f-66c9-49d3-be97-33b399bd1477" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="03">
      <SiteRef LocationOID="1001" />
      <StudyEventData StudyEventOID="VISIT1" StudyEventRepeatKey="VISIT1[1]">
        <FormData FormOID="VISIT" FormRepeatKey="1">
          <ItemGroupData ItemGroupOID="VISIT">
            <ItemData ItemOID="VISIT.VISITTM" TransactionType="Upsert">
              <AuditRecord>
                <UserRef UserOID="systemuser" />
                <LocationRef LocationOID="1001" />
                <DateTimeStamp>2014-08-13T12:20:35</DateTimeStamp>
                <ReasonForChange />
                <SourceID>6434490</SourceID>
              </AuditRecord>
              <mdsol:Query QueryRepeatKey="5606" Value="Data is required.  Please complete." Status="Open" Recipient="Site from System" />
            </ItemData>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
  </ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("QueryOpen", sc.subcategory)
        self.assertEqual(5606, sc.query.repeat_key)
        self.assertEqual("Data is required.  Please complete.", sc.query.value)
        self.assertEqual("Open", sc.query.status)
        self.assertEqual("Site from System", sc.query.recipient)

    def test_comment(self):
        """Test data entered with comment"""
        self.parse(
            """<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MEDICILLIN-RD7(DEMO)" MetaDataVersionOID="5" mdsol:AuditSubCategoryName="CommentAdd">
    <SubjectData SubjectKey="038f41bb-47bf-4776-8190-aaf442246f51" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="10001001">
      <SiteRef LocationOID="1000" />
      <StudyEventData StudyEventOID="SCREEN" StudyEventRepeatKey="SCREEN[1]">
        <FormData FormOID="SERCHOL" FormRepeatKey="1">
          <ItemGroupData ItemGroupOID="SERCHOL">
            <ItemData ItemOID="SERCHOL.CHOL_GROUP" TransactionType="Upsert">
              <AuditRecord>
                <UserRef UserOID="systemuser" />
                <LocationRef LocationOID="1000" />
                <DateTimeStamp>2013-08-26T19:11:01</DateTimeStamp>
                <ReasonForChange />
                <SourceID>47697</SourceID>
              </AuditRecord>
              <mdsol:Comment CommentRepeatKey="2" TransactionType="Insert" Value="This subject is now able to be randomized. Please navigate to the Randomization form to randomize the subject." />
            </ItemData>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
  </ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("CommentAdd", sc.subcategory)
        self.assertEqual(2, sc.comment.repeat_key)
        com = "This subject is now able to be randomized. Please navigate to the Randomization form to randomize the subject."
        self.assertEqual(com, sc.comment.value)
        self.assertEqual("Insert", sc.comment.transaction_type)

    def test_instance_name_changed(self):
        """ObjectNameChanged subcategory for changes of Instance names"""
        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MEDICILLIN-RD7(DEMO)" MetaDataVersionOID="5" mdsol:AuditSubCategoryName="ObjectNameChanged">
    <SubjectData SubjectKey="038f41bb-47bf-4776-8190-aaf442246f51" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="10001001">
      <SiteRef LocationOID="1000" />
      <StudyEventData StudyEventOID="UNSCHEDULED" StudyEventRepeatKey="UNSCHEDULED[1]" TransactionType="Upsert" mdsol:InstanceName="Unscheduled Visit 22 Aug 2013">
        <AuditRecord>
          <UserRef UserOID="systemuser" />
          <LocationRef LocationOID="1000" />
          <DateTimeStamp>2013-08-26T21:09:25</DateTimeStamp>
          <ReasonForChange />
          <SourceID>47976</SourceID>
        </AuditRecord>
      </StudyEventData>
    </SubjectData>
  </ClinicalData></ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("ObjectNameChanged", sc.subcategory)
        self.assertEqual("Upsert", sc.event.transaction_type)
        self.assertEqual("UNSCHEDULED[1]", sc.event.repeat_key)
        self.assertEqual("Unscheduled Visit 22 Aug 2013", sc.event.instance_name)
        self.assertIsNone(sc.event.instance_overdue)

    def test_datapage_name_changed(self):
        """ObjectNameChanged subcategory for changes of datapage names"""
        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MEDICILLIN-RD7(DEMO)" MetaDataVersionOID="5" mdsol:AuditSubCategoryName="ObjectNameChanged">
    <SubjectData SubjectKey="038f41bb-47bf-4776-8190-aaf442246f51" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="10001001">
      <SiteRef LocationOID="1000" />
      <StudyEventData StudyEventOID="UNSCHEDULED" StudyEventRepeatKey="UNSCHEDULED[1]">
        <FormData FormOID="VS" FormRepeatKey="1" TransactionType="Upsert" mdsol:DataPageName="Vital signs">
            <AuditRecord>
              <UserRef UserOID="systemuser" />
              <LocationRef LocationOID="1000" />
              <DateTimeStamp>2013-08-26T21:09:25</DateTimeStamp>
              <ReasonForChange />
              <SourceID>47976</SourceID>
            </AuditRecord>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData></ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("ObjectNameChanged", sc.subcategory)
        self.assertEqual("Upsert", sc.form.transaction_type)
        self.assertEqual(1, sc.form.repeat_key)
        self.assertEqual("Vital signs", sc.form.datapage_name)

    def test_instance_overdue(self):
        """When instance overdue date is set"""
        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MEDICILLIN-RD7(DEMO)" MetaDataVersionOID="5" mdsol:AuditSubCategoryName="InstanceOverdue">
    <SubjectData SubjectKey="038f41bb-47bf-4776-8190-aaf442246f51" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="10001001">
      <SiteRef LocationOID="1000" />
      <StudyEventData StudyEventOID="WEEK_03" StudyEventRepeatKey="WEEK_03[1]" TransactionType="Upsert" mdsol:InstanceOverdue="9/3/2013 12:00:00 AM">
        <AuditRecord>
          <UserRef UserOID="systemuser" />
          <LocationRef LocationOID="1000" />
          <DateTimeStamp>2013-08-26T20:04:59</DateTimeStamp>
          <ReasonForChange />
          <SourceID>47815</SourceID>
        </AuditRecord>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
</ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("InstanceOverdue", sc.subcategory)
        self.assertEqual("Upsert", sc.event.transaction_type)
        self.assertEqual("WEEK_03[1]", sc.event.repeat_key)
        self.assertEqual("9/3/2013 12:00:00 AM", sc.event.instance_overdue)
        self.assertIsNone(sc.event.instance_name)

    def test_protocol_deviation(self):
        """Protocol deviation creation"""
        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MEDICILLIN-RD7(DEMO)" MetaDataVersionOID="5" mdsol:AuditSubCategoryName="CreatePD">
    <SubjectData SubjectKey="162685e7-7445-46e2-9178-04276fbcfc92" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="10161001">
      <SiteRef LocationOID="1016" />
      <StudyEventData StudyEventOID="WEEK_01" StudyEventRepeatKey="WEEK_01[1]">
        <FormData FormOID="VS" FormRepeatKey="1">
          <ItemGroupData ItemGroupOID="VS">
            <ItemData ItemOID="VS.VSBMI" TransactionType="Upsert">
              <AuditRecord>
                <UserRef UserOID="systemuser" />
                <LocationRef LocationOID="1016" />
                <DateTimeStamp>2013-10-04T20:57:39</DateTimeStamp>
                <ReasonForChange />
                <SourceID>80966</SourceID>
              </AuditRecord>
              <mdsol:ProtocolDeviation TransactionType="Insert" Value="Body Mass Index is out of the prescribed range of LT 20 or GT 30" Status="Open" Class="Incl/Excl Criteria not met" Code="Deviation" ProtocolDeviationRepeatKey="218" />
            </ItemData>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
</ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("CreatePD", sc.subcategory)
        self.assertEqual("Insert", sc.protocol_deviation.transaction_type)
        self.assertEqual(
            "Body Mass Index is out of the prescribed range of LT 20 or GT 30",
            sc.protocol_deviation.value,
        )
        self.assertEqual("Open", sc.protocol_deviation.status)
        self.assertEqual("Incl/Excl Criteria not met", sc.protocol_deviation.klass)
        self.assertEqual(218, sc.protocol_deviation.repeat_key)
        self.assertEqual("Deviation", sc.protocol_deviation.code)

    def test_review(self):
        """Test for reviews"""
        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MEDICILLIN-RD7(DEMO)" MetaDataVersionOID="5" mdsol:AuditSubCategoryName="Review">
    <SubjectData SubjectKey="7d6d3179-2da2-4521-9068-3618f8f71e10" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="10081003">
      <SiteRef LocationOID="1008" />
      <StudyEventData StudyEventOID="SCREEN" StudyEventRepeatKey="SCREEN[1]">
        <FormData FormOID="VS" FormRepeatKey="1">
          <ItemGroupData ItemGroupOID="VS">
            <ItemData ItemOID="VS.TEMP" TransactionType="Upsert">
              <AuditRecord>
                <UserRef UserOID="jdustin.cdm@gmail.com" />
                <LocationRef LocationOID="1008" />
                <DateTimeStamp>2013-10-29T17:46:20</DateTimeStamp>
                <ReasonForChange>Data Management</ReasonForChange>
                <SourceID>134487</SourceID>
              </AuditRecord>
              <mdsol:Review GroupName="Data Management" Reviewed="Yes" />
            </ItemData>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData> </ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("Review", sc.subcategory)
        self.assertTrue(sc.review.reviewed)
        self.assertEqual("Data Management", sc.review.group_name)

    def test_verify(self):
        """Test data verification"""
        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MEDICILLIN-RD7(DEMO)" MetaDataVersionOID="5" mdsol:AuditSubCategoryName="Verify">
    <SubjectData SubjectKey="b5fb9054-0ba7-428a-8b82-678f25e346c0" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="10101001">
      <SiteRef LocationOID="1010" />
      <StudyEventData StudyEventOID="SCREEN" StudyEventRepeatKey="SCREEN[1]">
        <FormData FormOID="VISIT" FormRepeatKey="1">
          <ItemGroupData ItemGroupOID="VISIT">
            <ItemData ItemOID="VISIT.DTC" TransactionType="Upsert" mdsol:Verify="Yes">
              <AuditRecord>
                <UserRef UserOID="walbrocra" />
                <LocationRef LocationOID="1010" />
                <DateTimeStamp>2013-09-10T21:00:35</DateTimeStamp>
                <ReasonForChange />
                <SourceID>59545</SourceID>
              </AuditRecord>
            </ItemData>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>  </ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual(True, sc.item.verify)

    def test_signature(self):
        """Test signatures"""
        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MEDICILLIN-RD7(DEMO)" MetaDataVersionOID="5" mdsol:AuditSubCategoryName="ValidESigCredential">
    <SubjectData SubjectKey="c8717df8-cb50-4150-88c6-a6395c6653f5" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="10011001">
      <SiteRef LocationOID="1001" />
      <StudyEventData StudyEventOID="SCREEN" StudyEventRepeatKey="SCREEN[1]">
        <FormData FormOID="IE" FormRepeatKey="1" TransactionType="Upsert">
          <AuditRecord>
            <UserRef UserOID="mwissner.INV@gmail.com" />
            <LocationRef LocationOID="1001" />
            <DateTimeStamp>2013-08-29T16:11:31</DateTimeStamp>
            <ReasonForChange />
            <SourceID>52222</SourceID>
          </AuditRecord>
          <Signature>
            <UserRef UserOID="mwissner.INV@gmail.com" />
            <LocationRef LocationOID="1001" />
            <SignatureRef SignatureOID="5" />
            <DateTimeStamp>2013-08-29T16:11:31</DateTimeStamp>
          </Signature>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
    </ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("ValidESigCredential", sc.subcategory)
        self.assertEqual("5", sc.signature.oid)
        self.assertEqual("1001", sc.signature.location_oid)
        self.assertEqual("mwissner.INV@gmail.com", sc.signature.user_oid)
        self.assertEqual(
            datetime.datetime(2013, 8, 29, 16, 11, 31), sc.signature.datetimestamp
        )

    def test_form_datapage_id(self):
        """Test data entered with folder refs, reasons for change etc"""

        self.parse(
            u"""<?xml version="1.0" encoding="UTF-8"?>
<ODM xmlns="http://www.cdisc.org/ns/odm/v1.3" xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata" ODMVersion="1.3" FileType="Transactional" FileOID="4d690eda-4f08-48d1-af26-3bab40f6118f" CreationDateTime="2014-11-04T16:37:05">
  <ClinicalData StudyOID="MOVE-2014(DEV)" MetaDataVersionOID="2867" mdsol:AuditSubCategoryName="EnteredWithChangeCode">
    <SubjectData SubjectKey="538bdc4d-78b7-4ff9-a59c-3d13c8d8380b" mdsol:SubjectKeyType="SubjectUUID" mdsol:SubjectName="01">
      <SiteRef LocationOID="1001" />
      <StudyEventData StudyEventOID="VISIT1" StudyEventRepeatKey="VISIT1[1]">
        <FormData FormOID="VISIT" FormRepeatKey="1" mdsol:DataPageId="123">
          <ItemGroupData ItemGroupOID="VISIT">
            <ItemData ItemOID="VISIT.VISITDAT" TransactionType="Upsert" Value="7 Aug 2014">
              <AuditRecord>
                <UserRef UserOID="isparks" />
                <LocationRef LocationOID="1001" />
                <DateTimeStamp>2014-08-13T10:53:57</DateTimeStamp>
                <ReasonForChange>Data Entry Error</ReasonForChange>
                <SourceID>6434227</SourceID>
              </AuditRecord>
            </ItemData>
          </ItemGroupData>
        </FormData>
      </StudyEventData>
    </SubjectData>
  </ClinicalData>
  </ODM>""".encode(
                "ascii"
            )
        )

        sc = self.context

        self.assertEqual("EnteredWithChangeCode", sc.subcategory)
        self.assertEqual("VISIT1[1]", sc.event.repeat_key)
        self.assertEqual("Upsert", sc.item.transaction_type)
        self.assertEqual("VISIT", sc.form.oid)
        self.assertEqual(123, sc.form.datapage_id)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

import unittest

import pytest
from rwslib.rws_requests.odm_adapter import SignatureDefinitionsRequest, UsersRequest, \
    SitesMetadataRequest, VersionFoldersRequest, AuditRecordsRequest, LabAnalyteRangesRequest
from urllib.parse import quote


class TestSignatureDefinitionsRequest(unittest.TestCase):
    def create_request_object(self, project_name):
        t = SignatureDefinitionsRequest(project_name=project_name)
        return t

    def test_sig_def_requests(self):
        """Create a SignatureDefinitionsRequest"""
        project_name = "Mediflex(Prod)"
        t = self.create_request_object(project_name)
        self.assertEqual('datasets/Signatures.odm?studyid=%s' % quote(project_name),
                         t.url_path())


class TestUsersRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name="Mediflex",
                              environment="Dev",
                              location_oid=None):
        t = UsersRequest(project_name=project_name,
                         environment_name=environment,
                         location_oid=location_oid)
        return t

    def test_users_request(self):
        """Create a UsersRequest"""
        t = self.create_request_object()
        self.assertEqual('datasets/Users.odm?studyoid=%s' % quote(t.studyoid),
                         t.url_path())
        t = self.create_request_object(location_oid="1002")
        self.assertTrue('locationoid=%s' % quote("1002") in t.url_path())
        self.assertTrue('studyoid=%s' % quote(t.studyoid) in t.url_path())


class TestSitesMetadataRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name="Mediflex",
                              environment_name="Dev"):
        t = SitesMetadataRequest(project_name=project_name,
                                 environment_name=environment_name)
        return t

    def test_create_a_sites_metadata_request(self):
        """Create a SitesMetadataRequest"""
        t = self.create_request_object()
        self.assertEqual('datasets/Sites.odm?studyoid=%s' % quote(t.studyoid),
                         t.url_path())
        t = self.create_request_object(project_name=None,
                                       environment_name=None)
        self.assertEqual('datasets/Sites.odm',
                         t.url_path())

    def test_error_cases(self):
        """Create a SitesMetadataRequest, with error cases"""
        with self.assertRaises(AttributeError) as exc:
            t = self.create_request_object(environment_name=None)
        self.assertEqual("environment_name cannot be empty if project_name is set",
                         str(exc.exception))
        with self.assertRaises(AttributeError) as exc:
            t = self.create_request_object(environment_name='')
        self.assertEqual("environment_name cannot be empty if project_name is set",
                         str(exc.exception))
        with self.assertRaises(AttributeError) as exc:
            t = self.create_request_object(project_name=None)
        self.assertEqual("project_name cannot be empty if environment_name is set",
                         str(exc.exception))
        with self.assertRaises(AttributeError) as exc:
            t = self.create_request_object(project_name='')
        self.assertEqual("project_name cannot be empty if environment_name is set",
                         str(exc.exception))


class TestVersionFoldersRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name="Mediflex",
                              environment_name="Dev"):
        t = VersionFoldersRequest(project_name=project_name,
                                  environment_name=environment_name)
        return t

    def test_happy_case(self):
        """we can create a VersionFoldersRequest"""
        t = self.create_request_object()
        self.assertEqual("datasets/VersionFolders.odm?studyoid=%s" % quote(t.studyoid),
                         t.url_path())


class TestAuditRecordsRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name="Mediflex",
                              environment_name="Dev",
                              startid=1,
                              per_page=100):
        t = AuditRecordsRequest(project_name=project_name,
                                environment_name=environment_name,
                                startid=startid,
                                per_page=per_page)
        return t

    def test_create_audit_records_request(self):
        """We can create an AuditRecordsRequest"""
        t = self.create_request_object()
        self.assertTrue('datasets/ClinicalAuditRecords.odm' in t.url_path())
        self.assertTrue('studyoid=%s' % quote(t.studyoid) in t.url_path())
        self.assertTrue('startid=1' in t.url_path())
        self.assertTrue('per_page=100' in t.url_path())
        t = self.create_request_object(startid=2569, per_page=50)
        self.assertTrue('startid=2569' in t.url_path())
        self.assertTrue('per_page=50' in t.url_path())


def test_clinical_audit_request_mode_allowed():
    """We can create an AuditRecordsRequest"""
    for allowed_mode in ("default", "all", "enhanced"):
        t = AuditRecordsRequest(project_name="Mediflex",
                                environment_name="Dev",
                                startid=1,
                                per_page=100,
                                mode=allowed_mode)
        assert 'datasets/ClinicalAuditRecords.odm' in t.url_path()
        assert f'mode={allowed_mode}' in t.url_path()


def test_clinical_audit_request_mode_not_allowed():
    """We can create an AuditRecordsRequest"""
    for not_allowed_mode in ("potato", "pear", "guava"):
        with pytest.raises(ValueError):
            t = AuditRecordsRequest(project_name="Mediflex",
                                    environment_name="Dev",
                                    startid=1,
                                    per_page=100,
                                    mode=not_allowed_mode)


def test_clinical_audit_request_unicode():
    """We can create an AuditRecordsRequest"""
    t = AuditRecordsRequest(project_name="Mediflex",
                            environment_name="Dev",
                            startid=1,
                            per_page=100,
                            unicode=True)
    assert 'datasets/ClinicalAuditRecords.odm' in t.url_path()
    assert f'unicode=true' in t.url_path()


def test_clinical_audit_request_unicode_default():
    """We can create an AuditRecordsRequest"""
    t = AuditRecordsRequest(project_name="Mediflex",
                            environment_name="Dev",
                            startid=1,
                            per_page=100)
    assert 'datasets/ClinicalAuditRecords.odm' in t.url_path()
    assert f'unicode' not in t.url_path()


def test_lab_ranges():
    """Test the creation of a LabRangesRequest"""
    t = LabAnalyteRangesRequest("Mediflex")
    assert t.url_path() == "datasets/LabAnalyteRanges.csv?studyoid=Mediflex"


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

from rwslib.builders.constants import ProtocolDeviationStatus, QueryStatusType
from rwslib.builders.clinicaldata import ItemData, MdsolProtocolDeviation, MdsolQuery
from ..common import obj_to_doc


class TestMdsolQuery(unittest.TestCase):
    """Test extension MdsolQuery"""

    def get_tested(self, **kwargs):
        defaults = dict(status=QueryStatusType.Open, value="Data missing", query_repeat_key=123,
                          recipient="Site from System", requires_response=True)
        defaults.update(kwargs)
        return MdsolQuery(**defaults)

    def test_basic(self):
        tested = self.get_tested()
        self.assertEqual("Data missing", tested.value)
        self.assertEqual(123, tested.query_repeat_key)
        self.assertEqual(QueryStatusType.Open, tested.status)
        self.assertEqual("Site from System", tested.recipient)
        self.assertEqual(True, tested.requires_response)

    def test_builder(self):
        tested = self.get_tested()
        tested.response = "Done"
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:Query", doc.tag)
        self.assertEqual("Yes", doc.attrib['RequiresResponse'])
        self.assertEqual("Site from System", doc.attrib['Recipient'])
        self.assertEqual("123", doc.attrib['QueryRepeatKey'])
        self.assertEqual("Data missing", doc.attrib['Value'])
        self.assertEqual("Done", doc.attrib['Response'])

    def test_invalid_status_value(self):
        """Status must come from QueryStatusType"""
        with self.assertRaises(AttributeError):
            MdsolQuery(status='A test')

    def test_preceding_query_repeat_key(self):
        """Test that preceding query repeat key is passed through"""
        tested = self.get_tested(status=QueryStatusType.Open,
                                 preceding_query_repeat_key=11891,
                                 query_repeat_key=None,
                                 response=None)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:Query", doc.tag)
        self.assertEqual("Yes", doc.attrib['RequiresResponse'])
        self.assertEqual("Site from System", doc.attrib['Recipient'])
        self.assertIsNone(doc.attrib.get('QueryRepeatKey'))
        self.assertEqual("Data missing", doc.attrib['Value'])
        self.assertEqual("11891", doc.attrib['PrecedingQueryRepeatKey'])


class TestProtocolDeviation(unittest.TestCase):
    """Test extension MdsolProtocolDeviation"""

    def test_define_protocol_deviation(self):
        """Create a simple protocol deviation"""
        pd = MdsolProtocolDeviation(value="Deviated from Protocol",
                                    status=ProtocolDeviationStatus.Open,
                                    repeat_key=1)
        tested = obj_to_doc(pd)
        self.assertEqual("mdsol:ProtocolDeviation", tested.tag, "Unexpected Tag")
        self.assertEqual("Open", tested.attrib['Status'], "Status Key is missing")
        self.assertEqual("Deviated from Protocol", tested.get('Value'), "Value is missing")
        self.assertEqual(1, tested.get('ProtocolDeviationRepeatKey'), "ProtocolDeviationRepeatKey is missing")

    def test_define_protocol_deviation_with_class(self):
        """Create a simple protocol deviation with class and code"""
        pd = MdsolProtocolDeviation(value="Deviated from Protocol",
                                    status=ProtocolDeviationStatus.Open,
                                    repeat_key=1, code="E01", klass="Blargle")
        tested = obj_to_doc(pd)
        self.assertEqual("mdsol:ProtocolDeviation", tested.tag, "Unexpected Tag")
        self.assertEqual("Open", tested.attrib['Status'], "Status Key is missing")
        self.assertEqual("Deviated from Protocol", tested.get('Value'), "Value is missing")
        self.assertEqual("E01", tested.get('Code'), "Code is missing")
        self.assertEqual("Blargle", tested.get('Class'), "Class is missing")
        self.assertEqual(1, tested.get('ProtocolDeviationRepeatKey'), "ProtocolDeviationRepeatKey is missing")

    def test_define_protocol_deviation_with_transaction_type(self):
        """Create a simple protocol deviation with class and code and Transaction Type"""
        pd = MdsolProtocolDeviation(value="Deviated from Protocol",
                                    status=ProtocolDeviationStatus.Open,
                                    repeat_key=1, code="E01", klass="Blargle",
                                    transaction_type="Insert")
        tested = obj_to_doc(pd)
        self.assertEqual("mdsol:ProtocolDeviation", tested.tag, "Unexpected Tag")
        self.assertEqual("Open", tested.attrib['Status'], "Status Key is missing")
        self.assertEqual("Deviated from Protocol", tested.get('Value'), "Value is missing")
        self.assertEqual("E01", tested.get('Code'), "Code is missing")
        self.assertEqual("Blargle", tested.get('Class'), "Class is missing")
        self.assertEqual(1, tested.get('ProtocolDeviationRepeatKey'), "ProtocolDeviationRepeatKey is missing")
        self.assertEqual("Insert", tested.get('TransactionType'))

    def test_insert_pd_to_itemdata(self):
        """Create a simple protocol deviation with class and code and Transaction Type"""
        test = ItemData("FIXOID", "Fix me")(MdsolProtocolDeviation(value="Deviated from Protocol",
                                                                   status=ProtocolDeviationStatus.Open,
                                                                   repeat_key=1, code="E01", klass="Blargle",
                                                                   transaction_type="Insert"))
        tested = obj_to_doc(test)
        self.assertEqual(1, len(list(tested)))
        self.assertEqual('mdsol:ProtocolDeviation', list(tested)[0].tag)
        test << MdsolProtocolDeviation(value="Deviated from Protocol",
                                       status=ProtocolDeviationStatus.Open,
                                       repeat_key=2, code="E01", klass="Blargle",
                                       transaction_type="Insert")
        tested = obj_to_doc(test)
        self.assertEqual(2, len(list(tested)))
        self.assertEqual('mdsol:ProtocolDeviation', list(tested)[1].tag)
        self.assertEqual(2, list(tested)[1].get('ProtocolDeviationRepeatKey'))

    def test_define_protocol_deviation_with_errors(self):
        """Validate the entry"""
        with self.assertRaises(ValueError) as exc:
            pd = MdsolProtocolDeviation(value="Deviated from Protocol",
                                        status="Wigwam",
                                        repeat_key=1, code="E01", klass="Blargle")
        self.assertEqual("Status Wigwam is not a valid ProtocolDeviationStatus", str(exc.exception))
        with self.assertRaises(ValueError) as exc:
            pd = MdsolProtocolDeviation(value="Deviated from Protocol",
                                        status=ProtocolDeviationStatus.Open,
                                        repeat_key="no repeats", code="E01", klass="Blargle")
        self.assertEqual("RepeatKey should be an integer, not no repeats", str(exc.exception))



# -*- coding: utf-8 -*-

__author__ = 'glow'

from xml.etree import cElementTree as ET


def obj_to_doc(obj, *args, **kwargs):
    """Convert an object to am XML document object
    :rtype: xml.etree.ElementTree.Element
    """
    builder = ET.TreeBuilder()
    obj.build(builder, *args, **kwargs)
    return builder.close()



# -*- coding: utf-8 -*-

__author__ = 'glow'

import unittest

from six.moves.urllib_parse import quote

from rwslib.rws_requests.biostats_gateway import check_dataset_format, DATASET_FORMATS, \
    dataset_format_to_extension, CVMetaDataRequest, FormDataRequest, MetaDataRequest, \
    ProjectMetaDataRequest, ViewMetaDataRequest, CommentDataRequest, ProtocolDeviationsRequest, \
    DataDictionariesRequest


class TestCheckDatasetFormat(unittest.TestCase):
    def test_good_cases(self):
        """Good happy case for check_dataset_format"""
        try:
            check_dataset_format("CSV")
            check_dataset_format("XML")
            check_dataset_format("csv")
            check_dataset_format("xml")
        except ValueError:
            self.fail()

    def test_poor_cases(self):
        """Bad formats for check_dataset_format"""
        for nonsense in ['xls', 'dat', 'xslx']:
            with self.assertRaises(ValueError) as exc:
                check_dataset_format(nonsense)
            self.assertEqual("dataset_format is expected to be one of "
                             "%s. '%s' is not valid" % (', '.join(DATASET_FORMATS.keys()),
                                                        nonsense),
                             str(exc.exception))


class TestDatasetFormatToExtension(unittest.TestCase):
    def test_good_cases(self):
        """Good happy case for check_dataset_format"""
        for dsf, format_ in DATASET_FORMATS.items():
            match = dataset_format_to_extension(dsf)
            self.assertEqual(format_, match)

    def test_poor_cases(self):
        """Bad formats for check_dataset_format"""
        for nonsense in ['xls', 'dat', 'xslx']:
            with self.assertRaises(ValueError) as exc:
                dataset_format_to_extension(nonsense)
            self.assertEqual("dataset_format is expected to be one of "
                             "%s. '%s' is not valid" % (', '.join(DATASET_FORMATS.keys()),
                                                        nonsense),
                             str(exc.exception))


class TestCVMetaDataRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name,
                              environment_name,
                              versionitem=None,
                              rawsuffix=None,
                              codelistsuffix=None,
                              decodesuffix=None):
        t = CVMetaDataRequest(project_name,
                              environment_name,
                              versionitem=versionitem,
                              rawsuffix=rawsuffix,
                              codelistsuffix=codelistsuffix,
                              decodesuffix=decodesuffix)
        return t

    def test_create_cv_metadata_request(self):
        """Create a CVMetaDataRequest"""
        t = self.create_request_object("Mediflex", "Dev")
        self.assertEqual("Mediflex(Dev)", t.studyname_environment())
        self.assertEqual("studies/Mediflex(Dev)/datasets/metadata/regular", t.url_path())


class TestFormDataRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name,
                              environment_name,
                              dataset_type,
                              form_oid,
                              start=None,
                              dataset_format="csv"):
        t = FormDataRequest(project_name,
                            environment_name,
                            dataset_type,
                            form_oid,
                            start=start,
                            dataset_format=dataset_format)
        return t

    def test_successful_configuration(self):
        """We can successfully request FormData"""
        t = self.create_request_object("Mediflex", "Dev", "raw", "DM")
        self.assertEqual('studies/Mediflex(Dev)/datasets/raw/DM.csv', t.url_path())
        t = self.create_request_object("Mediflex", "Dev", "raw", "DM", dataset_format="xml")
        self.assertEqual('studies/Mediflex(Dev)/datasets/raw/DM', t.url_path())


class TestMetaDataRequest(unittest.TestCase):
    def create_request_object(self,
                              dataset_format="csv"):
        t = MetaDataRequest(dataset_format=dataset_format)
        return t

    def test_successful_configuration(self):
        """We can successfully request MetaData"""
        t = self.create_request_object()
        self.assertEqual('datasets/ClinicalViewMetadata.csv', t.url_path())
        t = self.create_request_object(dataset_format="xml")
        self.assertEqual('datasets/ClinicalViewMetadata', t.url_path())


class TestProjectMetaDataRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name="Mediflex",
                              dataset_format="csv"):
        t = ProjectMetaDataRequest(project_name=project_name,
                                   dataset_format=dataset_format)
        return t

    def test_successful_configuration(self):
        """We can successfully request ProjectMetaData"""
        t = self.create_request_object()
        self.assertEqual('datasets/ClinicalViewMetadata.csv?ProjectName=Mediflex', t.url_path())
        t = self.create_request_object(dataset_format="xml")
        self.assertEqual('datasets/ClinicalViewMetadata?ProjectName=Mediflex', t.url_path())


class TestViewMetaDataRequest(unittest.TestCase):
    def create_request_object(self,
                              view_name="DM",
                              dataset_format="csv"):
        t = ViewMetaDataRequest(view_name=view_name,
                                dataset_format=dataset_format)
        return t

    def test_successful_configuration(self):
        """We can successfully request ViewMetaData"""
        t = self.create_request_object()
        self.assertEqual('datasets/ClinicalViewMetadata.csv?ViewName=DM', t.url_path())
        t = self.create_request_object(dataset_format="xml")
        self.assertEqual('datasets/ClinicalViewMetadata?ViewName=DM', t.url_path())


class TestCommentDataRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name,
                              environment_name,
                              dataset_format="csv"):
        t = CommentDataRequest(project_name=project_name,
                               environment_name=environment_name,
                                dataset_format=dataset_format)
        return t

    def test_successful_configuration(self):
        """We can successfully request CommentData"""
        t = self.create_request_object('Mediflex', 'Dev')
        self.assertEqual('datasets/SDTMComments.csv?studyid=%s' % quote(t.studyname_environment()),
                         t.url_path())
        t = self.create_request_object('Mediflex', 'Dev', dataset_format="xml")
        self.assertEqual('datasets/SDTMComments?studyid=%s' % quote(t.studyname_environment()),
                         t.url_path())


class TestProtocolDeviationsRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name,
                              environment_name,
                              dataset_format="csv"):
        t = ProtocolDeviationsRequest(project_name=project_name,
                                      environment_name=environment_name,
                                      dataset_format=dataset_format)
        return t

    def test_successful_configuration(self):
        """We can successfully request ProtocolDeviations"""
        t = self.create_request_object('Mediflex', 'Dev')
        self.assertEqual('datasets/SDTMProtocolDeviations.csv?studyid=%s' % quote(t.studyname_environment()),
                         t.url_path())
        t = self.create_request_object('Mediflex', 'Dev', dataset_format="xml")
        self.assertEqual('datasets/SDTMProtocolDeviations?studyid=%s' % quote(t.studyname_environment()),
                         t.url_path())


class TestDataDictionariesRequest(unittest.TestCase):
    def create_request_object(self,
                              project_name,
                              environment_name,
                              dataset_format="csv"):
        t = DataDictionariesRequest(project_name=project_name,
                                    environment_name=environment_name,
                                    dataset_format=dataset_format)
        return t

    def test_successful_configuration(self):
        """We can successfully request ProtocolDeviations"""
        t = self.create_request_object('Mediflex', 'Dev')
        self.assertEqual('datasets/SDTMDataDictionaries.csv?studyid=%s' % quote(t.studyname_environment()),
                         t.url_path())
        t = self.create_request_object('Mediflex', 'Dev', dataset_format="xml")
        self.assertEqual('datasets/SDTMDataDictionaries?studyid=%s' % quote(t.studyname_environment()),
                         t.url_path())



# -*- coding: utf-8 -*-
__author__ = 'isparks'

import logging
from typing import Optional

from six.moves.urllib.parse import urlparse, parse_qs

from rwslib.extras.audit_event.parser import parse
from rwslib.rws_requests.odm_adapter import AuditRecordsRequest


class ODMAdapter(object):
    """A self-contained data fetcher and parser using a RWSConnection and an event class provided by the user"""
    def __init__(self, rws_connection, study, environment, eventer, mode: Optional[str] = None):
        self.rws_connection = rws_connection
        self.eventer = eventer
        self.study = study
        self.environment = environment
        self.mode = mode
        self.start_id = 0

    def get_next_start_id(self):
        """If link for next result set has been passed, extract it and get the next set start id"""
        link = self.rws_connection.last_result.links.get("next", None)
        if link:
            link = link['url']
            p = urlparse(link)
            start_id = int(parse_qs(p.query)['startid'][0])
            return start_id

        return None

    def run(self, start_id=0, max_pages=-1, per_page=1000, **kwargs):
        page = 0
        self.start_id = start_id
        while max_pages == -1 or (page < max_pages):

            req = AuditRecordsRequest(self.study, self.environment, startid=self.start_id, per_page=per_page,
                                      mode=self.mode)
            try:
                # Get the ODM data
                odm = self.rws_connection.send_request(req, **kwargs)
                # Check if we were passed the next startid
                # Need to do this immediately because subsequent parsing might include other calls to RWS
                self.start_id = self.get_next_start_id()
                # Send it for parsing
                parse(odm, self.eventer)
                page += 1
            except Exception as e:
                logging.error(e.message)

            if not self.start_id:
                break
